from typing import Callable, List, Tuple
import numpy as np
import logging
from numpy.linalg.linalg import norm as nrm
from yupi.trajectory import Trajectory
from yupi.transformations import subsample
from yupi._checkers import (
    _check_uniform_time_spaced,
    _check_same_dt,
    _check_same_dim,
    _check_exact_dim,
    _check_same_t
)


_COMPONENT_NUM = { 'x':0, 'y':1, 'z':2 }

def _parse_collect_key(value: str) -> Callable:
    original_value = value
    is_delta = value.startswith('d')
    value = value[1:] if is_delta else value
    if len(value) == 0 or value[0] not in ['r', 'v', 'a']:
        raise ValueError(f"Unknown key '{original_value}'")

    vector_name = value[0]
    component = value[1:]
    is_norm = component == 'n'
    if not is_norm:
        if component.isnumeric():
            component = int(component)
        elif component in ['x','y','z']:
            component = _COMPONENT_NUM[component]
        elif component == '':
            component = -1
        else:
            raise ValueError(f"Unknown key '{original_value}'")
    else:
        component = -1

    def _key(traj: Trajectory, delta=True, norm=True):
        data = None
        if vector_name == 'r':
            data = traj.r
        elif vector_name == 'v':
            data = traj.v
        else:
            data = traj.ang

        if is_delta and delta:
            data = data.delta
        if is_norm and norm:
            return data.norm
        if component == -1:
            return data
        return data.component(component)

    return _key, is_delta, is_norm


def collect_at(trajs: List[Trajectory], key: str, step: int = None,
               time: float = None, warnings: bool = True) -> np.ndarray:
    is_step = step is not None
    is_time = time is not None
    if is_step + is_time == 0:
        is_step = True
        step = 1
    if is_step + is_time == 2:
        raise ValueError("You can not set 'step' and 'time' parameter at the "
                        "same time")

    key = _parse_collect_key(key)[0]
    data = np.zeros(len(trajs))
    for i, traj in enumerate(trajs):
        step = int(time / traj.dt) if is_time else step
        if warnings and step >= len(traj):
            logging.warning(f"Trajectory {i} with id={traj.traj_id} is "
                            f"shorten than {step} samples")
            continue

        data[i] = key(traj)[step]
    return data


def collect(trajs: List[Trajectory], key: str, lag_step: int = None,
            lag_time: float = None, concat: bool = True,
            warnings: bool = True) -> np.ndarray:
    """
    Collects the information requested by the key parameter from an
    ensemble of trajectories.

    If the key contains the ``delta`` signature (e.g. 'drx') then
    the extracted data will be subsampled first according ``lag_step``
    or ``lag_time`` parameters. If none of this parameters is given
    then it is assumed ``lag_step = 1``.

    Parameters
    ----------
    trajs : List[Trajectory]
        Group of trajectories.
    key : str
        Describes what information will be extracted from the
        trajectories.
    lag_step : int, optional
        Index distance between samples, by default Nonee.
    lag_time : float, optional
        Time distance between samples, by default None.
    concat : bool, optional
        If true each trajectory stracted data will be concatenated in
        a single array, by default True.

    Returns
    -------
    np.ndarray
        Collected data.

    Raises
    ------
    ValueError
        If ``lag_step`` and ``lag_time`` are given at the same time.
    ValueError
        If the given lag is grater than one of the trajectories length.
    """

    is_step = lag_step is not None
    is_time = lag_time is not None
    if is_step + is_time == 0:
        is_step = True
        lag_step = 1
    if is_step + is_time == 2:
        raise ValueError("You can not set 'lag_step' and 'lag_time' parameter "
                         "at the same time")

    key, is_delta, is_norm = _parse_collect_key(key)

    data = []

    if not is_delta:
        for i, traj in enumerate(trajs):
            step = int(lag_time / traj.dt) if is_time else int(lag_step)

            if warnings and step >= len(traj):
                logging.warning(f"Trajectory {i} with id={traj.traj_id} is "
                                f"shorten than {step} samples")
                continue

            current_data = key(traj)[::step]

            if concat:
                data.extend(current_data)
            else:
                data.append(current_data)
        return np.array(data)

    vectors = [key(traj, delta=False, norm=False) for traj in trajs]

    for i, vec in enumerate(vectors):
        traj = trajs[i]
        step = int(lag_time / traj.dt) if is_time else int(lag_step)

        if warnings and step >= len(traj):
            logging.warning(f"Trajectory {i} with id={traj.traj_id} is "
                            f"shorten than {step} samples")
            continue

        current_data = vec[step:] - vec[:-step]

        if is_norm:
            current_data = np.linalg.norm(current_data, axis=1)

        if concat:
            data.extend(current_data)
        else:
            data.append(current_data)
    return np.array(data)


@_check_same_dt
@_check_exact_dim(2)
@_check_uniform_time_spaced
def turning_angles_ensemble(trajs: List[Trajectory], accumulate=False,
                            degrees=False, centered=False,
                            wrap=True) -> np.ndarray:
    """
    Return a concatenation of all the turning angles that forms
    a set of trajectories.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.
    accumulate : bool, optional
        If True, turning angles are measured with respect to an axis
        define by the initial velocity (i.e., angles between initial
        and current velocity). Otherwise, relative turning angles
        are computed (i.e., angles between succesive velocity vectors).
        By default False.
    degrees : bool, optional
        If True, angles are given in degrees. Otherwise, the units
        are radians. By default False.
    centered : bool, optional
        If True, angles are wrapped on the interval ``[-pi, pi]``.
        Otherwise, the interval ``[0, 2*pi]`` is chosen. By default
        False.

    Returns
    -------
    np.ndarray
        Concatenated array of turning angles for a list of Trajectory
        objects.
    """

    theta = [t.turning_angles(accumulate, degrees, centered, wrap) for t in trajs]
    return np.concatenate(theta)


@_check_same_dim
def speed_ensemble(trajs: List[Trajectory], step: int = 1) -> np.ndarray:
    """
    Estimate speeds of the list of trajectories, ``trajs``,
    by computing displacements according to a certain sample
    frequency given by ``step``.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.
    step : int
        Numer of sample points.

    Returns
    -------
    np.array
        Concatenated array of speeds.
    """

    trajs_ = [subsample(traj, step) for traj in trajs]
    return np.concatenate([traj.v.norm for traj in trajs_])


@_check_same_t
def msd_ensemble(trajs: List[Trajectory]) -> np.ndarray:
    """
    Compute the square displacements for every Trajectory object
    stored in ``trajs`` as the square of the current position vector
    that has been subtracted the initial position.

    Trajectories should have the same length.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.

    Returns
    -------
    np.ndarray
        Array of square displacements with shape ``(n, N)``, where
        ``n`` is the total number of time steps and ``N`` the number
        of trajectories.
    """

    _msd = []
    for traj in trajs:
        # Position vectors
        r = traj.r

        # Square displacements
        r_2 = (r - r[0])**2            # Square coordinates
        r2 = np.sum(r_2, axis=1)       # Square distances
        _msd.append(r2)                 # Append square distances

    # Transpose to have time/trials as first/second axis
    _msd = np.transpose(_msd)
    return _msd


@_check_same_dt
@_check_uniform_time_spaced
def msd_time(trajs: List[Trajectory], lag: int) -> np.ndarray:
    """
    Estimate the mean square displacement for every Trajectory
    object stored in ``trajs`` as the average of the square of
    dispacement vectors as a function of the lag time.

    This is a convenience estimator specially when trajectories
    do not have equal lengths.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.
    lag : int
        Number of steps that multiplied by ``dt`` defines the lag
        time.

    Returns
    -------
    np.ndarray
        Array of mean square displacements with shape ``(lag, N)``,
        where ``N`` the number of trajectories.
    """

    _msd = []
    for traj in trajs:
        # Position vectors
        r = traj.r

        # Compute msd for a single trajectory
        current_msd = np.empty(lag)
        for lag_ in range(1, lag + 1):
            # Lag displacement vectors
            dr = r[lag_:] - r[:-lag_]
            # Lag displacement
            dr2 = np.sum(dr**2, axis=1)
            # Averaging over a single realization
            current_msd[lag_ - 1] = np.mean(dr2)

        # Append all square displacements
        _msd.append(current_msd)

    # Transpose to have time/trials as first/second axis
    _msd = np.transpose(_msd)
    return _msd


@_check_same_dim
def msd(trajs: List[Trajectory], time_avg: bool = True,
        lag: int = None) -> Tuple[np.ndarray, np.ndarray]:
    """
    Estimate the mean square displacement of the list of Trajectory
    objects, ``trajs``, providing the options of averaging over the
    ensemble of realizations or over time.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.
    time_avg : bool, optional
        If True, mean square displacement is estimated averaging over
        time. Otherwise, an ensemble average will be performed and all
        Trajectory objects will have to have the same length. By default
        True.
    lag : None, optional
        If None, ``time_avg`` should be set to ``False`` indicating
        ensemble average. Otherwise, ``lag`` is taken as the number
        of steps that multiplied by ``dt`` defines the lag time. By
        default None.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        Tuple containing the array of mean square displacements and
        the array of standard deviations.
    """

    if not time_avg:
        _msd = msd_ensemble(trajs)   # Ensemble average
    elif lag is None:
        raise ValueError("You must set 'lag' param if 'time_avg' is True")
    else:
        _msd = msd_time(trajs, lag)  # Time average

    msd_mean = np.mean(_msd, axis=1)  # Mean
    msd_std = np.std(_msd, axis=1)    # Standard deviation
    return msd_mean, msd_std


@_check_same_t
def vacf_ensemble(trajs: List[Trajectory]) -> np.ndarray:
    """
    Compute the pair-wise dot product between initial and current
    velocity vectors for every Trajectory object stored in ``trajs``.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.

    Returns
    -------
    np.ndarray
        Array of velocity dot products with shape ``(n, N)``, where
        ``n`` is the total number of time steps and ``N`` the number
        of trajectories.
    """

    _vacf = []
    for traj in trajs:
        # Cartesian velocity components
        v = traj.v

        # Pair-wise dot product between velocities at t0 and t
        v0_dot_v = np.sum(v[0] * v, axis=1)

        # Append all veloctiy dot products
        _vacf.append(v0_dot_v)

    # Transpose to have time/trials as first/second axis
    _vacf = np.transpose(_vacf)
    return _vacf


@_check_same_dt
@_check_uniform_time_spaced
def vacf_time(trajs: List[Trajectory], lag: int) -> np.ndarray:
    """
    Estimate the velocity autocorrelation function for every
    Trajectory object stored in ``trajs`` as the average of the
    dot product between velocity vectors that are distant a certain
    lag time.

    This is a convenience estimator specially when trajectories do
    not have equal lengths.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.
    lag : int
        Number of steps that multiplied by ``dt`` defines the lag
        time.

    Returns
    -------
    np.ndarray
        Array of velocity autocorrelation function with shape
        ``(lag, N)``, where ``N`` is the number of trajectories.
    """

    _vacf = []
    for traj in trajs:
        # Cartesian velocity components
        v = traj.v

        # Compute vacf for a single trajectory
        current_vacf = np.empty(lag)
        for lag_ in range(1, lag + 1):
            # Multiply components given lag
            v1v2 = v[:-lag_] * v[lag_:]

            # Dot product for a given lag time
            v1_dot_v2 = np.sum(v1v2, axis=1)

            # Averaging over a single realization
            current_vacf[lag_ - 1] = np.mean(v1_dot_v2)

        # Append the vacf for a every single realization
        _vacf.append(current_vacf)

    # Aranspose to have time/trials as first/second axis
    _vacf = np.transpose(_vacf)
    return _vacf


@_check_same_dim
def vacf(trajs: List[Trajectory], time_avg: bool = True,
        lag: int = None) -> Tuple[np.ndarray, np.ndarray]:
    """
    Estimate the velocity autocorrelation function of the list of
    Trajectory objects, ``trajs``, providing the options of averaging
    over the ensemble of realizations or over time.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.
    time_avg : bool, optional
        If True, velocity autocorrelation function is estimated
        averaging over time. Otherwise, an ensemble average will be
        performed and all Trajectory objects will have to have the
        same length. By default True.
    lag : int, optional
        If None, ``time_avg`` should be set to ``False`` indicating
        ensemble average. Otherwise, ``lag`` is taken as the number
        of steps that multiplied by ``dt`` defines the lag time.
        By default None.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        Tuple containing the array of velocity autocorrelation function
        and the array of standard deviations.
    """

    if not time_avg:
        _vacf = vacf_ensemble(trajs)   # Ensemble average
    elif lag is None:
        raise ValueError("You must set 'lag' param if 'time_avg' is True")
    else:
        _vacf = vacf_time(trajs, lag)  # Time average

    vacf_mean = np.mean(_vacf, axis=1)  # Mean
    vacf_std = np.std(_vacf, axis=1)    # Standard deviation
    return vacf_mean, vacf_std


def _kurtosis(arr):
    """
    Compute the kurtosis of the array, `arr`.

    If `arr` is not a one-dimensional array, it should
    be a horizontal collection of column vectors.

    Parameters
    ----------
    arr : np.adarray
        Data for which the kurtosis is calculated.

    Returns
    -------
    float
        Kurtosis of the data set.
    """

    arr = np.squeeze(arr)

    # ONE-DIMENSIONAL CASE
    if len(arr.shape) == 1:
        # Subtract the mean position at every time instant
        arr_zm = arr - arr.mean()

        # Second and fourth central moments averaging
        # over repetitions
        m2 = np.mean(arr_zm**2)
        m4 = np.mean(arr_zm**4)

        # Compute kurtosis for those cases in which the
        # second moment is different from zero
        if m2 == 0:
            return 0
        kurt = m4 / m2**2
        return kurt

    # MULTIDIMENSIONAL CASE
    # arr should have shape (dim, trials)
    # (i.e., a horizontal sequence of column vectors)

    # Subtract the mean position
    arr_zm = arr - arr.mean(1)[:,None]

    try:
        # Inverse of the estimated covariance matrix
        cov_inv = np.linalg.inv(np.cov(arr))
    except np.linalg.LinAlgError:
        # Exception for the case of singular matrices
        return 0

    # Kurtosis definition for multivariate r.v.'s
    k = np.sum(arr_zm * (cov_inv @ arr_zm), axis=0)
    kurt = np.mean(k**2)

    return kurt

@_check_same_t
def kurtosis_ensemble(trajs: List[Trajectory]) -> np.ndarray:
    """Estimate kurtosis as a function of time of the
    list of Trajectory objects, ``trajs``. The average
    is perform over the ensemble of realizations.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.

    Returns
    -------
    np.ndarray
        Kurtosis at every time instant.
    """

    # Get ensemble positions where axis 0/1/2 are
    # in the order trials/time/dim
    r = [traj.r for traj in trajs]

    # Set trials as the last axis
    r = np.moveaxis(r, 0, 2)

    # Compute kurtosis at every time instant (loop over time)
    kurt = [_kurtosis(r_) for r_ in r]

    return np.array(kurt)


@_check_same_dt
@_check_uniform_time_spaced
def kurtosis_time(trajs: List[Trajectory], lag: int) -> np.ndarray:
    """
    Estimate the kurtosis for every Trajectory object stored
    in ``trajs``.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.
    lag : int
        Number of steps that multiplied by ``dt`` defines the lag
        time.

    Returns
    -------
    np.ndarray
        Array of velocity autocorrelation function with shape
        ``(lag, N)``, where ``N`` is the number of trajectories.
    """

    kurt = []
    for traj in trajs:
        current_kurt = np.empty(lag)
        for lag_ in range(lag):
            try:
                dr = traj.r[lag_:] - traj.r[:-lag_]
            except ValueError:
                current_kurt[lag_] = 0
                continue
            current_kurt[lag_] = _kurtosis(dr.T)
        kurt.append(current_kurt)
    kurt = np.transpose(kurt)
    return kurt


@_check_same_dim
def kurtosis(trajs: List[Trajectory], time_avg: bool = True,
             lag: int = None) -> Tuple[np.ndarray, np.ndarray]:
    """
    Estimate the kurtosis of the list of Trajectory objects, ``trajs``,
    providing the options of averaging over the ensemble of realizations
    or over time.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.
    time_avg : bool, optional
        If True, kurtosis is estimated averaging over time. Otherwise,
        an ensemble average will be performed and all Trajectory objects
        will have to have the same length. By default True.
    lag : int, optional
        If None, ``time_avg`` should be set to ``False`` indicating
        ensemble average. Otherwise, ``lag`` is taken as the number
        of steps that multiplied by ``dt`` defines the lag time.
        By default None.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        Tuple containgin the kurtosis and the standar deviations.
    """

    if not time_avg:
        return kurtosis_ensemble(trajs), None
    elif lag is None:
        raise ValueError("You must set 'lag' param if 'time_avg' is True")

    kurt = kurtosis_time(trajs, lag)
    kurt_mean = np.mean(kurt, axis=1)
    kurt_std = np.std(kurt, axis=1)
    return kurt_mean, kurt_std


@_check_same_dim
def kurtosis_reference(trajs: List[Trajectory]) -> float:
    """Get the sampled kurtosis for the case of
    ``len(trajs)`` trajectories whose position
    vectors are normally distributed.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input trajectories.

    Returns
    -------
    float
        Value of kurtosis.
    """

    d = trajs[0].dim
    n = len(trajs)
    kurt = d * (d + 2)
    if n == 1:
        return kurt
    return kurt * (n - 1) / (n + 1)


@_check_same_dt
@_check_uniform_time_spaced
def psd(trajs: List[Trajectory], lag: int, omega: bool = False) -> np.ndarray:
    """
    Estimate the power spectral density of a list of Trajectory object
    as the Fourier transform of its velocity autocorrelation function.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.
    lag : int
        Number of steps that multiplied by ``dt`` defines the lag
        time.
    omega: bool
        If True, return (psd, omega), where omega is the array of
        angular frequencies. Otherwise, just the power spectrum is
        returned. By default False.

    Returns
    -------
    np.ndarray
        Power spectral density.
    """

    _vacf = vacf_time(trajs, lag)
    ft = np.fft.fft(_vacf, axis=0) * trajs[0].dt
    ft = np.fft.fftshift(ft)
    ft_abs = np.abs(ft)
    ft_mean = np.mean(ft_abs, axis=1)
    ft_std = np.std(ft_abs, axis=1)

    if not omega:
        return ft_mean, ft_std, None

    omega_ = 2*np.pi * np.fft.fftfreq(lag, trajs[0].dt)
    omega_ = np.fft.fftshift(omega_)
    return ft_mean, ft_std, omega_



"""
This module contains a set of functions capable of applying
transformations to trajectories such as filtering, resampling, etc.

All the resources of this module should be imported directly
from ``yupi.transormations``.
"""

from yupi.transformations._resamplers import subsample
from yupi.transformations._filters import exp_convolutional_filter
from yupi.transformations._transformations import add_moving_FoR

__all__ = [
    'subsample',
    'exp_convolutional_filter',
    'add_moving_FoR'
]



import itertools
from typing import List
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
from yupi.graphics._style import (
    LIGHT_YELLOW,
    LIGHT_BLUE,
    LIGHT_ORANGE,
    GREEN,
    LIGHT_GREEN,
    RED,
    YUPI_COLORS,
    YUPI_LIGHT_COLORS,
    _plot_basic_properties
)

def _validate_units(units):
    return '' if units is None else f' [{units}]'

@_plot_basic_properties
def plot_hist(values: np.ndarray, **kwargs):
    """
    Plot a histogram of the given

    Parameters
    ----------
    values : np.ndarray
        Input values
    """

    if 'color' not in kwargs:
        kwargs['color'] = YUPI_LIGHT_COLORS[0]
    if 'ec' not in kwargs:
        kwargs['ec'] = (0,0,0,0.6)
    plt.hist(values, **kwargs)

@_plot_basic_properties
def plot_hists(values_list: List[np.ndarray], kwargs_list: List[dict] = None,
               labels: List[str] = None, filled: bool = False,
               **general_kwargs):
    """
    Plot several histograms given a collection of values.

    Parameters
    ----------
    values_list : List[np.ndarray]
        Collection of values.
    kwargs_list : List[dict], optional
        kwargs of each plot, by default []

        If given, the length must be the same as the length of
        ``values``.
    """

    if kwargs_list and len(kwargs_list) != len(values_list):
        raise ValueError("The length of 'kwargs_list' must be equals the "
                         "length of 'values_list'")

    kwargs_list = [{}] * len(values_list)

    cycle = itertools.cycle(YUPI_COLORS)
    colors = [cycle.__next__() for _ in values_list]

    for i, vals in enumerate(values_list):
        color = colors[i]
        alpha = 0.3 if filled else 1
        kwargs = kwargs_list[i]
        kwargs = kwargs if len(kwargs) != 0 else general_kwargs
        lw = 1.5
        if labels is not None:
            kwargs['label'] = labels[i]
        if 'histtype' in kwargs:
            kwargs.pop('histtype')
        if 'color' in kwargs:
            color = kwargs['color']
        if 'lw' in kwargs:
            lw = kwargs['lw']
        if 'alpha' not in kwargs:
            kwargs['alpha'] = alpha
        alpha = kwargs.pop('alpha', alpha)

        plt.hist(vals, histtype='step', color=color, lw=lw, **kwargs)

        if filled:
            kwargs.pop('label', None)
            plt.hist(vals, histtype='stepfilled', color=color, alpha=alpha,
                     **kwargs)



def plot_velocity_hist(v, show: bool = True, units: str = 'm/s', **kwargs):
    """Plot a histogram of the array of velocities ``v``.

    Parameters
    ----------
    v : np.ndarray
        Velocity array.
    show : bool, optional
        If True, the plot is shown. By default True.
    units : string, optional
        Velocity units. By default 'm/s'.
    """

    if 'color' not in kwargs:
        kwargs['color'] = LIGHT_YELLOW

    if 'density' in kwargs:
        kwargs.pop('density')

    units = _validate_units(units)

    plt.hist(v, ec=(0,0,0,0.6), density=True, **kwargs)
    plt.xlabel(f'speed{units}')
    plt.ylabel('pdf')
    plt.grid()
    plt.gca().set_axisbelow(True)

    if show:
        plt.show()


def plot_angles_hist(ang, bins, show: bool = True, ax=None):
    """Plot a histogram of the array of angles ``ang``.

    Parameters
    ----------
    ang : np.ndarray
        Array of angles.
    bins: int
        Number of histogram bins.
    show : bool, optional
        If True, the plot is shown. By default True.
    ax : matplotlib.axes._subplots.AxesSubplot, optional
        Axes to plot. By default None.
    """

    if ax is None:
        ax = plt.gca(projection='polar')
    plt.hist(ang, bins, density=True, ec=(0,0,0,0.6), color=LIGHT_BLUE)
    ax.set_theta_zero_location('N')
    ax.set_rlabel_position(135)
    ax.set_axisbelow(True)
    plt.xlabel('turning angles pdf')
    if show:
        plt.show()


def plot_msd(msd, msd_std, dt, lag, x_units: str = 's', y_units: str = 'm^2/s',
             show=True):
    """Plot Mean Square Displacement.

    Parameters
    ----------
    msd : np.ndarray
        Mean square displacement array.
    msd_std : np.ndarray
        Standard deviation.
    dt : float
        Trajectories time step.
    lag : int, optional
        Lag time.
    x_units : str, optional
        Units of the time axes.
    y_units : str, optional
        Units of the MSD axes.
    show : bool, optional
        If True, the plot is shown. By default True.
    """

    x_units = _validate_units(x_units)
    y_units = _validate_units(y_units)

    lag_t_msd = dt * np.arange(lag)
    plt.plot(lag_t_msd, msd, color='.2')
    upper_bound = msd + msd_std
    lower_bound = msd - msd_std
    plt.fill_between(lag_t_msd, upper_bound, lower_bound, color=LIGHT_ORANGE)
    plt.xlabel(f'lag time{x_units}')
    plt.ylabel(r'$\mathrm{msd \;' + y_units + '}$')
    plt.grid()
    if show:
        plt.show()


def plot_kurtosis(kurtosis, dt=None, t_array=None, kurtosis_ref: float = None,
                  units: str = 's', show=True):
    """Plot kurtosis.

    Parameters
    ----------
    kurtosis : np.adarray
        Kurtosis array.
    dt : float
        Trajectories time step.
    t_array : np.ndarray, optional
        Array of time instants that match with every value in
        ``kurtosis``. By default None.
    kurtosis_ref : float, optional
        The value of kurtosis for a gaussian.
    units : str, optional
        Units of the time axes.
    show : bool, optional
        If True, the plot is shown. By default True.
    """

    units = _validate_units(units)

    if dt:
        t_array = np.linspace(0, dt*len(kurtosis), len(kurtosis))
    if t_array is not None:
        plt.plot(t_array, kurtosis, color=GREEN)

        if kurtosis_ref is not None:
            bound_1 = kurtosis
            bound_2 = [kurtosis_ref]*len(t_array)
            plt.fill_between(t_array, bound_1, bound_2, color=LIGHT_GREEN)
        plt.xlabel(f'time{units}')
    else:
        plt.plot(kurtosis)

    plt.ylabel('kurtosis')
    plt.grid()
    if show:
        plt.show()


def plot_vacf(vacf, dt, lag, x_units: str = 's', y_units: str = '(m/s)^2',
              show=True):
    """Plot Velocity Autocorrelation Function.

    Parameters
    ----------
    vacf : np.ndarray
        Velocity autocorrelation function array.
    dt : float
        Trajectories time step.
    lag : int
        Number of steps that multiplied by ``dt`` defines the lag
        time.
    x_units : str, optional
        Units of the time axes.
    y_units : str, optional
        Units of the VACF axes.
    show : bool, optional
        If True, the plot is shown. By default True.
    """

    x_units = _validate_units(x_units)
    y_units = _validate_units(y_units)

    lag_t_vacf = dt * np.arange(lag)

    plt.plot(lag_t_vacf, vacf, '.', color=RED)
    plt.xlabel(f'lag time{x_units}')
    plt.ylabel(r'$\mathrm{vacf \;' + y_units + '}$')
    plt.grid()

    ax = plt.gca()

    inset_axes(ax, width='60%', height='60%', bbox_to_anchor=(0, 0, 1, 1),
               bbox_transform=ax.transAxes, loc='upper right')

    plt.plot(lag_t_vacf, vacf, '.', color=RED)
    plt.yscale('log')
    plt.grid()

    if show:
        plt.show()


def plot_psd(psd_mean, omega, psd_std=None, show=True):
    """Plot the Power Spectral Density.

    Parameters
    ----------
    psd_mean : np.ndarray
        Power spectral density array.
    omega : np.ndarray
        Array of angular frequencies.
    psd_std : np.ndarray, optional
        Standard deviation of the power spectrum.
        By default None.
    show : bool, optional
        If True, the plot is shown. By default True.
    """

    plt.plot(omega, psd_mean, label='psd')
    if psd_std is not None:
        plt.fill_between(omega, psd_mean - psd_std, psd_mean + psd_std,
            alpha=.3, label='psd_std')
    plt.xscale('log')
    plt.yscale('log')
    plt.legend()
    if show:
        plt.show()



import logging
from typing import Callable
from pathlib import Path
import cv2
import numpy as np
from yupi.tracking.algorithms import TrackingAlgorithm, _resize_frame
from yupi.transformations._affine_estimator import _get_affine
from yupi.trajectory import Trajectory
from yupi.transformations import add_moving_FoR
from yupi.tracking.undistorters import Undistorter


class ROI():
    """
    Region of interest.

    Region that can be tracked by the algorithms throughout the sequence
    of image frames.

    Parameters
    ----------
    size : tuple of float
        Size of the region of interest.

        If both tuple's values are grater than 1 then they are rounded
        and taken as pixels. Otherwise, if both values are less than 1,
        the size is taken relative to the video frame size.
    init_mode : str, optional
        ROI's initialization mode, by default 'manual'.

        Defines the way ROI initial position is setted.

        The ``init_mode`` parameter can be manual or center. These
        modes are stored in ``ROI.MANUAL_INIT_MODE`` and
        ``ROI.CENTER_INIT_MODE``.
    scale : float, optional
        Scale of the sample frame to set ROI initial position if
        ``init_method`` is set to ``'manual'``, by default 1.

    Attributes
    ----------
    width : float
        Width of the ROI.

        If the width value is between 0 and 1 then this is taken
        relative to the frames. Otherwise it is a rounded value and
        taken as pixels.
    height : float
        Height of the ROI.

        If the height value is between 0 and 1 then this is taken
        relative to the frames. Otherwise it is a rounded value and
        taken as pixels.
    init_mode : str
        ROI's initialization mode.
    scale : float
        Scale of the sample frame to set ROI initial position if
        ``init_method`` is set to ``'manual'``.

    Examples
    --------
    >>> ROI((120, 120), ROI.MANUAL_INIT_MODE)
    ROI: size=(120, 120) init_mode=manual scale=1

    Raises
    ------
    ValueError
        If any size value is negative.
    ValueError
        If one of the size value is grater than 1 and the other is less
        than 1.
    ValueError
        If ROI initialization mode is neither ``'manual'`` or
        ``'center'``.
    """

    MANUAL_INIT_MODE = 'manual'
    """Manual initialization mode for the ROI"""
    CENTER_INIT_MODE = 'center'
    """Center initialization mode for the ROI"""

    def __init__(self, size: tuple, init_mode: str = MANUAL_INIT_MODE,
                 scale: float = 1):

        if size[0] <= 0 or size[1] <= 0:
            raise ValueError("ROI's size values must be positives")

        # TODO: check for a more pythonic way to do this comprobation
        if (size[0] < 1 and size[1] > 1) or \
           (size[0] > 1 and size[1] < 1):
            raise ValueError('Size values must be between 0 and 1 both or '
                             'integers greater than 0 both')

        if init_mode not in (ROI.CENTER_INIT_MODE, ROI.MANUAL_INIT_MODE):
            raise ValueError(f"ROI '{init_mode}' initialization mode unknown")

        if scale < 0:
            raise ValueError('ROI scale must be non negative')

        self.width, self.height = size
        self.init_mode = init_mode
        self.scale = scale
        self.__prev_cXY = None, None
        self.__cXY = None, None
        self.__global_height, self.__global_width = None, None

    # This repr could change
    def __repr__(self):
        return 'ROI: size=({}, {}) init_mode={} scale={}' \
            .format(self.width, self.height, self.init_mode, self.scale)

    def _recenter(self, centroid: tuple) -> tuple:
        """
        Recenters ROI position.

        Parameters
        ----------
        centroid : tuple of int
            New center of the ROI.

        Returns
        -------
        cX, cY : int
            Center of the ROI.
        """

        # Get the centroid refered to the roi
        if centroid is not None:
            cX_roi, cY_roi = centroid

            # Get the centroid refered to the full image
            cX = self.__prev_cXY[0] - int(self.width/2) + cX_roi
            cY = self.__prev_cXY[1] - int(self.height/2) + cY_roi

            cX = min(cX, self.__global_width)
            cX = max(cX, 0)
            cY = min(cY, self.__global_height)
            cY = max(cY, 0)

            self.__cXY = cX, cY

        else:
            self.__cXY = self.__prev_cXY

    def _get_bounds(self, prev: bool = False) -> tuple:
        """
        ROI's bounds.

        Calculates the ROI's bounds according to its center, width,
        height and the global bounds.

        Parameters
        ----------
        prev : bool
            Use previous roi center instead of current

        Returns
        -------
        xmin : int
            Mnimum bound on X axis.
        xmax : int
            Maximun bound on X axis.
        ymin : int
            Mnimum bound on Y axis.
        ymax : int
            Maximum bound on Y axis.
        """

        if prev:
            cX, cY = self.__prev_cXY
        else:
            cX, cY = self.__cXY

        half_width, half_height = int(self.width/2), int(self.height/2)
        xmin = max(cX - half_width, 0)
        xmax = min(cX + half_width, self.__global_width)
        ymin = max(cY - half_height, 0)
        ymax = min(cY + half_height, self.__global_height)
        return xmin, xmax, ymin, ymax

    def _center_init(self, frame: np.ndarray) -> tuple:
        """
        Initialize ROI using center initialization mode.

        Parameters:
        frame : np.ndarray
            Frame used as reference to initialize ROI position at its
            center.

        Returns
        -------
        tuple of int
            Center of the ROI.
        """

        self.__global_height, self.__global_width = frame.shape[:2]
        self.__cXY = self.__global_width // 2, self.__global_height // 2
        return self.__cXY

    # TODO: check for 'win2_name' utility. Maybe it should be 'ROI' as
    # Default so there is no need to pass it as a parameter
    def _manual_init(self, frame: np.ndarray, name: str) -> tuple:
        """
        Initialize ROI using manual initialization mode.

        Parameter
        ---------
        frame : np.ndarray
            Frame used as reference to initialize ROI position manually.
        name : str
            Name of the tracking object.

        Returns
        -------
        tuple of int
            Center of the ROI.
        """

        win1_name = 'Click on the center of {} to init roi'.format(name)

        self.__global_height, self.__global_width = frame.shape[:2]

        frame_ = _resize_frame(frame, scale=self.scale)
        cv2.imshow(win1_name, frame_)

        # Callback handler to manually set the roi
        def on_click(event, x, y, flags, param):

            if event == cv2.EVENT_LBUTTONDOWN:
                # Global roi center coordinates
                self.__cXY = int(x / self.scale), int(y / self.scale)

                # Copy of true frame and its resized version
                img_ = frame_.copy()

                # Draw a circle in the selected pixel
                cv2.circle(img_, (x, y), 3, (0, 255, 255), 1)
                xmin, xmax, ymin, ymax = self._get_bounds()
                pt1 = (int(xmin * self.scale), int(ymin * self.scale))
                pt2 = (int(xmax * self.scale), int(ymax * self.scale))
                cv2.rectangle(img_, pt1, pt2, (0, 255, 255), 1)
                cv2.imshow(win1_name, img_)

                logging.info('ROI initialized, now press any key to continue')

        cv2.setMouseCallback(win1_name, on_click)
        cv2.waitKey(0)
        return self.__cXY

    def _check_roi_init(self, name: str) -> bool:
        """
        Checks for ROI initialization.

        Parameter
        ---------
        name : str
            Name of the tracking object.

        Returns
        -------
        bool
            Whether or not the ROI is initialized.
        """

        if not self.__prev_cXY[0]:
            logging.error(f'ROI was not initialized in {name}')
            return False
        else:
            cv2.destroyAllWindows()
            logging.info(f'ROI initialized in {name}')
            return True

    def _initialize(self, name: str, first_frame: np.ndarray) -> bool:
        """
        Initialize ROI.

        Parameters
        ----------
        name : str
            Name of the tracking object.
        first_frame : np.ndarray
            First frame of the video.

            If ROI's initialization mode is set to ``'manual'`` this
            frame will be shown to select the tracking object center.

        Returns
        -------
        bool
            Whether or not the ROI was initialized.
        """

        h, w = first_frame.shape[:2]
        if self.width <= 1:
            self.width *= w
        if self.height <= 1:
            self.height *= h

        # Initialize ROI coordinates manually by user input
        if self.init_mode == ROI.MANUAL_INIT_MODE:
            self.__cXY = self._manual_init(first_frame, name)
            self.__prev_cXY = self.__cXY
        else:
            self.__cXY = self._center_init(first_frame)
            self.__prev_cXY = self.__cXY

        return self._check_roi_init(name)

    def _crop(self, frame: np.ndarray, prev: bool = False) -> np.ndarray:
        """
        Crops a frame according to the ROI's bounds.

        Parameters
        ----------
        frame : np.ndarray
            Frame that will be cropped.
        prev : bool
            Use previous roi center instead of current

        Returns
        -------
        window : np.ndarray
            Cropped part of the frame.
        """

        self.__global_height, self.__global_width = frame.shape[:2]
        # Bounds of the roi
        xmin, xmax, ymin, ymax = self._get_bounds(prev)
        window = frame[ymin:ymax, xmin:xmax, :]
        return window.copy()


class ObjectTracker():
    """
    Tracks an object inside a ROI according to a tracking algorithm.

    Parameters
    ----------
    name : str
        Name of the tracked object.
    algorithm : TrackingAlgorithm
        Algorithm used to track the object.
    roi : ROI
        Region of interest where the object will be tracked.
    preprocessing : Callable[[np.ndarray], np.ndarray], optional
        Preprocessing function aplied to the frame before being used by
        the algorithm.

    Attributes
    ----------
    name : str
        Name of the tracked object.
    algorithm : TrackingAlgorithm
        Algorithm used to track the object.
    roi : ROI
        Region of interest where the object will be tracked.
    history : list of tuple
        ROI's position in every frame of the video.
    preprocessing : Callable[[np.ndarray], np.ndarray]
        Preprocessing function aplied to the frame before being used by
        the algorithm.
    """

    def __init__(self, name: str, algorithm: TrackingAlgorithm, roi: ROI,
                 preprocessing: Callable[[np.ndarray], np.ndarray] = None):
        self.name = name
        self.roi = roi
        self.history = []
        self.algorithm = algorithm
        self.preprocessing = preprocessing
        self.mask = None

    def _init_roi(self, frame: np.ndarray) -> bool:
        return self.roi._initialize(self.name, frame)

    def _track(self, frame: np.ndarray) -> tuple:
        """
        Tracks the center of the object.

        Given a new frame, the center of the object inside the ROI is
        recalculated using the selected algorithm.

        Parameters
        ----------
        frame : np.ndarray
            Frame used by the algorithm to detect the tracked object's
            new center.
        """

        # Get only the ROI from the current frame
        roi_bound = self.roi._get_bounds()

        # Detect the object using the tracking algorithm
        self.mask, centroid = self.algorithm.detect(frame,
                                                    roi_bound,
                                                    self.preprocessing)

        # Update the roi center using current ant coordinates
        self.roi._recenter(centroid)

        # Update data
        self.history.append(self.roi._ROI__cXY)


class CameraTracker():
    """
    Tracks the camera movement.

    Parameters
    ----------
    roi : ROI
        Region of interest where the background changes will be
        detected.

    Attributes
    ----------
    roi : ROI
        Region of interest where the background changes will be
        detected.
    history : list of tuple
        ROI's position in every frame of the video.
    """

    def __init__(self, roi: ROI):
        self.history = []
        self.mse = []
        self.roi = roi
        self.features = None

    def _init_roi(self, prev_frame: np.ndarray) -> bool:
        return self.roi._initialize('Camera', prev_frame)

    # Track the floor
    def _track(self, prev_frame: np.ndarray, frame: np.ndarray,
               ignored_regions: list) -> bool:
        """
        Tracks the camera movements according to the changing background
        inside the ROI.

        Parameters
        ----------
        prev_frame, frame : np.ndarray
            Frames used to detect background movement.
        igonerd_regions : list of tuple
            Tracked object's boundaries.

            Tracked object's does not form part of the background so
            they should be ignored.

        Returns
        -------
        bool
            Whether or not good points were found or sucessfully
            tracked.
        """

        # Initialize a mask of what to track
        h, w = frame.shape[:2]
        mask = 255 * np.ones((h, w), dtype=np.uint8)

        # Mask pixeles inside every ROIs
        for x0, xf, y0, yf in ignored_regions:
            mask[y0:yf, x0:xf] = 0

        p_good, aff_params, err = _get_affine(
            img1=prev_frame,
            img2=frame,
            region=self.roi._get_bounds(),
            mask=mask
        )
        self.features = p_good[1:]

        if err is None:
            return False

        self.history.append(aff_params)
        self.mse.append(err)

        return True


class TrackingScenario():
    """
    Controls all the tracking process along the video.

    Parameters
    ----------
    object_trackers : list
        Trackers of all the objects.
    camera_tracker : CameraTracker
        Tracker used to detect camera movements, by default None.
    undistorter : Undistorter
        Undistorted used to correct each video frame, by default None.
    preview_scale : float
        Scale of the video preview, by default 1.0.
    auto_mode : bool
        If True the video is processed auomtically otherwise it's
        processed manually, by default True.

        If the video is processed manually, pressing ``ENTER`` key is
        necessary in every frame to continue.

        This mode can be changed in the middle of the processing by
        pressing ``M`` key.

    Attributes
    ----------
    object_trackers : list
        Trackers of all the objects
    camera_tracker : CameraTracker
        Tracker used to detect camera movements.
    undistorter : Undistorter
        Undistorted used to correct each video frame.
    preview_scale : float
        Scale of the video preview.
    auto_mode : bool
        If True the video is processed auomtically otherwise it's
        processed manually, by default True.

        If the video is processed manually, pressing ``Enter`` key is
        necessary in every frame to continue.

        This mode can be changed in the middle of the processing by
        pressing ``M`` key.
    """

    def __init__(self, object_trackers: list,
                 camera_tracker: CameraTracker = None,
                 undistorter: Undistorter = None,
                 preview_scale: float = 1,
                 auto_mode: bool = True):
        self.object_trackers = object_trackers
        self.camera_tracker = camera_tracker
        self.undistorter = undistorter
        self.preview_scale = preview_scale
        self.auto_mode = auto_mode
        self._enabled = True
        self._iteration_counter = 0

        self.video_path = None
        self.cap = None
        self.frame_count = None
        self.fps = None
        self.w = None
        self.h = None
        self.dim = None
        self.first_frame = None
        self.last_frame = None
        self.prev_frame = None

    def _digest_video_path(self, video_path):
        if not Path.exists(Path(video_path)):
            raise ValueError(f"Path '{video_path}' does not exists")
        self.video_path = video_path

        # Create capture object
        self.cap = cv2.VideoCapture(video_path)

        # Total number of frames in the video file
        self.frame_count = int(self.cap.get(cv2.CAP_PROP_FRAME_COUNT))

        # Frames per seconds
        self.fps = self.cap.get(cv2.CAP_PROP_FPS)

        # Frame width
        self.w = int(self.cap.get(cv2.CAP_PROP_FRAME_WIDTH))

        # Frame height
        self.h = int(self.cap.get(cv2.CAP_PROP_FRAME_HEIGHT))

        self.dim = (self.w, self.h)

        self.first_frame = 0

    def _undistort(self, frame):
        if self.undistorter:
            frame = self.undistorter.fix(frame)
        return frame

    def _show_frame(self, frame, show_frame_id=True):
        # CXY, region, features, frame_numb, mask
        frame = frame.copy()

        # Draw region in which features are detected
        if self.camera_tracker:
            x0, xf, y0, yf = self.camera_tracker.roi._get_bounds()

            cv2.putText(frame, 'Camera Tracking region', (x0+5, yf-5),
                        cv2.FONT_HERSHEY_COMPLEX_SMALL, 1.2, (0, 0, 255), 1,
                        cv2.LINE_AA)

            cv2.rectangle(frame, (x0, y0), (xf, yf), (0, 0, 255), 2)
            p2, p3 = self.camera_tracker.features
            # Draw detected and estimated features
            for p2_, p3_ in zip(p2, p3):
                x2, y2 = np.rint(p2_).astype(np.int32)
                x3, y3 = np.rint(p3_).astype(np.int32)

                cv2.circle(frame, (x2, y2), 3, (0, 0, 0), -1)
                cv2.circle(frame, (x3, y3), 3, (0, 255, 0), -1)

        for otrack in self.object_trackers:
            # TODO: Do this better:
            # Alter the blue channel in ant-related pixels
            window = otrack.roi._crop(frame, prev=True)
            if otrack.mask is not None:
                window[:, :, 0] = otrack.mask

            # Draw a point over the roi center and draw bounds
            x1, x2, y1, y2 = otrack.roi._get_bounds()
            cv2.circle(frame, otrack.roi._ROI__cXY, 5, (255, 255, 255), -1)
            cv2.rectangle(frame, (x1, y1), (x2, y2), (0, 255, 255), 2)
            cv2.putText(frame, otrack.name, (x1+5, y2-5),
                        cv2.FONT_HERSHEY_COMPLEX_SMALL, 1.2, (0, 255, 255), 1,
                        cv2.LINE_AA)

        if show_frame_id:
            h, w = frame.shape[:2]
            frame_id = self._iteration_counter + self.first_frame
            x_, y_ = .02, .05
            x, y = int(x_ * w), int(y_ * h)
            cv2.putText(frame, str(frame_id), (x, y),
                        cv2.FONT_HERSHEY_COMPLEX_SMALL, 1.2, (0, 255, 255), 1,
                        cv2.LINE_AA)

        frame = _resize_frame(frame, self.preview_scale)
        cv2.imshow('yupi processing window', frame)
        # Return frame

    def _first_iteration(self, start_frame):
        # Start processing frams at the given index
        if start_frame:
            self.first_frame = start_frame
            self.cap.set(cv2.CAP_PROP_POS_FRAMES, start_frame)

        # Capture the first frame to process
        _, prev_frame = self.cap.read()

        # Correct spherical distortion
        self.prev_frame = self._undistort(prev_frame)

        # Initialize the roi of all the trackers
        for otrack in self.object_trackers:
            retval = otrack._init_roi(self.prev_frame)
            if not retval:
                return retval

        # Initialize the region of the camera tracker
        if self.camera_tracker:
            self.camera_tracker._init_roi(self.prev_frame)

        # Increase the iteration counter
        self._iteration_counter += 1

        logging.info('All trackers were initialized')
        return True

    def _keyboard_controller(self):
        # Keyboard events
        wait_key = 0 if not self.auto_mode else 10

        k = cv2.waitKey(wait_key) & 0xff
        if k == ord('m'):
            self.auto_mode = not self.auto_mode

        elif k == ord('q'):
            self._enabled = False

        elif k == ord('e'):
            exit()

    def _regular_iteration(self):
        # Get current frame and ends the processing when no more frames are
        # detected

        frame_id = self._iteration_counter + self.first_frame
        if self.last_frame is not None and frame_id >= self.last_frame:
            return False, True


        ret, frame = self.cap.read()
        if not ret:
            logging.info('All frames were processed')
            return False, True

        # Correct spherical distortion
        frame = self._undistort(frame)

        # ROI Arrays of tracking objects
        roi_array = []

        # Track every object and save past and current ROIs
        for otrack in self.object_trackers:
            roi_array.append(otrack.roi._get_bounds())
            otrack._track(frame)
            roi_array.append(otrack.roi._get_bounds())

        if self.camera_tracker:
            ret = self.camera_tracker._track(self.prev_frame, frame,
                                             roi_array)

        if not ret:
            msg = f'CameraTracker - No matrix was estimated (Frame {frame_id})'
            logging.error(msg)
            return False, False

        # Display the full image with the ant in blue (TODO: Refactor this to
        # Make more general)
        self._show_frame(frame)

        for otrack in self.object_trackers:
            otrack.roi._ROI__prev_cXY = otrack.roi._ROI__cXY

        # Save current frame and ROI center as previous for next iteration
        self.prev_frame = frame.copy()

        # Call the keyboard controller to handle key interruptions
        self._keyboard_controller()

        # Increase the iteration counter
        self._iteration_counter += 1

        return True, False

    def _release_cap(self):
        self.cap.release()
        cv2.destroyAllWindows()

    def _tracker2trajectory(self, tracker, pix_per_m):
        dt = 1/self.fps
        traj_id = tracker.name
        x, y = map(list, zip(*tracker.history))
        x_arr = np.array(x) / pix_per_m
        y_arr = -1 * np.array(y) / pix_per_m
        return Trajectory(x=x_arr, y=y_arr, dt=dt, traj_id=traj_id)

    def _export_trajectories(self, pix_per_m):
        t_list = []
        # Extract camera reference
        if self.camera_tracker:
            affine_params = np.array(self.camera_tracker.history)
            theta, tx, ty, _ = affine_params.T
            tx, ty = tx / pix_per_m, ty / pix_per_m
            # Invert axis
            theta *= -1
            ty *= -1
            reference = theta, tx, ty
        # Output the trajectory of each tracker
        for otrack in self.object_trackers:
            t = self._tracker2trajectory(otrack, pix_per_m)
            if self.camera_tracker:
                t = add_moving_FoR(t, reference, new_traj_id=t.traj_id)
            t_list.append(t)
        return t_list

    def track(self, video_path: str, start_frame: int = 0,
              end_frame: int = None, pix_per_m: int = 1):
        """
        Starts the tracking process.

        Parameters
        ----------
        video_path : str
            Path of the video used to track the objects.
        start_frame : int, optional
            Initial frame in which starts the processing, by default 0.
        end_frame : int, optional
            Last frame being processed, if nothing is passed all frames
            until the end of the video will be processed, by default
            None.
        pix_per_m : int, optional
            Pixel per meters, by default 1.

            This value is used to readjuts the trajectories points to a
            real scale.

        Returns
        -------
        bool
            Whether or not the tracking process ended succefully.
        list
            List of all the trajectories extracted in the tracking
            process.
        """

        if end_frame is not None and end_frame > start_frame:
            self.last_frame = int(end_frame)

        self._digest_video_path(video_path)

        end = False
        if self._iteration_counter == 0:
            retval = self._first_iteration(start_frame)
            if not retval:
                return retval, None

        logging.info('Processing frames')
        while self._enabled:
            retval, end = self._regular_iteration()
            if not retval:
                break

        if end:
            retval = True

        self._release_cap()
        trajectories = self._export_trajectories(pix_per_m)
        return retval, trajectories



import itertools
import logging
import matplotlib.pyplot as plt
from typing import List, Union
from yupi import Trajectory
from yupi.graphics._style import YUPI_COLORS, LINE



def plot_2D(trajs: Union[List[Trajectory], Trajectory], line_style: str = LINE,
            title: str = None, legend: bool = True, show: bool = True,
            connected: bool = False, units: str = 'm', color = None, **kwargs):
    """
    Plot all the points of trajectories from ``trajs`` in a 2D plane.

    Parameters
    ----------
    trajs : Union[List[Trajectory], Trajectory]
        Input trajectories.
    line_style : str
        Type of the trajectory line to plot. It uses the matplotlib,
        notation, by default '-'.
    title : str, optional
        Title of the plot, by default None.
    legend : bool, optional
        If True, legend is shown. By default True.
    show : bool, optional
        If True, the plot is shown. By default True.
    connected : bool
        If True, all the trajectory points of same index will be,
        connected.

        If the trajectories do not have same length then the points
        will be connected until the shortest trajectory last index.
    color : str or tuple or list
        Defines the color of the trajectories, by default None.

        If color is of type ``str`` or ``tuple`` (rgb) then the color
        is applied to all trajectories. If color is of type ``list``
        then the trajectories take the color according to the index.

        If there are less colors than trajectories then the remaining
        trajectories are colored automatically (not with the same
        color).
    """

    if isinstance(trajs, Trajectory):
        trajs = [trajs]

    units = '' if units is None else f' [{units}]'

    cycle = itertools.cycle(YUPI_COLORS)
    colors = [cycle.__next__() for _ in trajs]

    if color is not None:
        if isinstance(color, (str, tuple)):
            kwargs['color'] = color
        elif isinstance(color, list):
            colors = color

    if connected:
        lengths = list(map(len, trajs))
        min_len = min(lengths)
        max_len = max(lengths)
        if min_len != max_len:
            logging.warning('Not all the trajectories have the same length.')
        for i in range(min_len):
            traj_points = [t[i] for t in trajs]
            traj_points.append(traj_points[0])
            for tp1, tp2 in zip(traj_points[:-1], traj_points[1:]):
                xs = [tp1.r[0], tp2.r[0]]
                ys = [tp1.r[1], tp2.r[1]]
                plt.plot(xs, ys, color=(0.2, 0.2, 0.2), linewidth=0.5)

    for i, t in enumerate(trajs):

        if t.dim != 2:
            logging.warning(f'Using plot_2D with a trajectory of {t.dim} dimensions.')
            logging.warning(f'Trajectory No. {i} with id {t.traj_id})')

        # Plotting
        x, y = t.r.x, t.r.y
        if colors is not None:
            if i < len(colors):
                kwargs['color'] = colors[i]
            else:
                kwargs.pop('color')
        traj_plot = plt.plot(x, y, line_style, **kwargs)
        color = traj_plot[-1].get_color()
        plt.plot(x[0], y[0], 'o', mfc='white', zorder=2,
                 label=f'{t.traj_id} initial position', color=color)
        plt.plot(x[-1], y[-1], 'o', mfc='white', zorder=2, color=color)
        plt.plot(x[-1], y[-1], 'o', alpha=.5,
                 label=f'{t.traj_id} final position', color=color)

        if legend:
            plt.legend()

        plt.title(title)
        plt.tick_params(direction='in')
        plt.axis('equal')
        plt.grid(True)
        plt.xlabel(f'x{units}')
        plt.ylabel(f'y{units}')

    if show:
        plt.show()



def plot_3D(trajs: Union[List[Trajectory], Trajectory], line_style: str = LINE,
            title: str = None, legend: bool = True, show: bool = True,
            connected: bool = False, units: str = 'm', color = None, **kwargs):
    """
    Plot all the points of trajectories from ``trajs`` in a 3D space.

    Parameters
    ----------
    trajs : Union[List[Trajectory], Trajectory]
        Input trajectories.
    line_style : str
        Type of the trajectory line to plot. It uses the matplotlib,
        notation, by default '-'.
    title : str, optional
        Title of the plot, by default None.
    legend : bool, optional
        If True, legend is shown. By default True.
    show : bool, optional
        If True, the plot is shown. By default True.
    connected : bool
        If True, all the trajectory points of same index will be,
        connected.

        If the trajectories do not have same length then the points
        will be connected until the shortest trajectory last index.
    color : str or tuple or list
        Defines the color of the trajectories, by default None.

        If color is of type ``str`` or ``tuple`` (rgb) then the color
        is applied to all trajectories. If color is of type ``list``
        then the trajectories take the color according to the index.

        If there are less colors than trajectories then the remaining
        trajectories are colored automatically (not with the same
        color).
    """

    if isinstance(trajs, Trajectory):
        trajs = [trajs]

    units = '' if units is None else f' [{units}]'

    cycle = itertools.cycle(YUPI_COLORS)
    colors = [cycle.__next__() for _ in trajs]

    if color is not None:
        if isinstance(color, (str, tuple)):
            kwargs['color'] = color
        elif isinstance(color, list):
            colors = color

    
    ax = plt.gca(projection='3d')

    if connected:
        lengths = list(map(len, trajs))
        min_len = min(lengths)
        max_len = max(lengths)
        if min_len != max_len:
            logging.warning('Not all the trajectories have the same length.')
        for i in range(min_len):
            traj_points = [t[i] for t in trajs]
            traj_points.append(traj_points[0])
            for tp1, tp2 in zip(traj_points[:-1], traj_points[1:]):
                xs = [tp1.r[0], tp2.r[0]]
                ys = [tp1.r[1], tp2.r[1]]
                zs = [tp1.r[2], tp2.r[2]]
                ax.plot(xs, ys, zs, color=(0.2, 0.2, 0.2), linewidth=0.5)

    for i, t in enumerate(trajs):

        if t.dim != 3:
            logging.warning(f'Using plot_2D with a trajectory of {t.dim} dimensions.')
            logging.warning(f'Trajectory No. {i} with id {t.traj_id})')
            

        # Plotting
        x, y, z = t.r.x, t.r.y, t.r.z

        if colors is not None:
            if i < len(colors):
                kwargs['color'] = colors[i]
            else:
                kwargs.pop('color')
        traj_plot = ax.plot(x, y, z, line_style, **kwargs)
        color = traj_plot[-1].get_color()

        ax.plot(x[0], y[0], z[0], 'o', mfc='white',
                 label=f'{t.traj_id} initial position', color=color)
   
        ax.plot(x[-1], y[-1], z[-1], 'o', mfc='white', color=color) 
        ax.plot(x[-1], y[-1], z[-1], 'o', alpha=.5,
                 label=f'{t.traj_id} final position', color=color)

        if legend:
            plt.legend()

        plt.title(title)
        plt.tick_params(direction='in')
        plt.grid(True)
        plt.xlabel(f'x{units}')
        plt.ylabel(f'y{units}')

    if show:
        plt.show()

    


from typing import Tuple
import numpy as np
from yupi import Trajectory
from yupi.transformations._affine_estimator import _affine_matrix


def add_moving_FoR(traj: Trajectory,
                   reference: Tuple[np.ndarray, np.ndarray, np.ndarray],
                   start_at_origin: bool = True, new_traj_id: str = None):
    """
    This function fuses the information of a trajectory with an
    external reference of the motion of the Frame of Reference
    (FoR).

    It allows to remap the information gathered in local SoRs
    to a more general FoR.

    Parameters
    ----------
    traj : Trajectory
        Input trajectory.
    reference : Tuple[np.ndarray,np.ndarray,np.ndarray]
        Angular and translational parameters of the form
        ``(ang:np.ndarray, tx:np.ndarray, ty:np.ndarray)`` that
        accounts for the orientation and displacement of the reference.
    start_at_origin : bool, optional
        If True, set initial position at the origin. By default True.

    Returns
    -------
    Trajectory
        Output trajectory in the lab frame of reference.
    """

    def affine2camera(theta, tx, ty):
        x_cl, y_cl, theta_cl = np.zeros((3, theta.size + 1))
        theta_cl[1:] = np.cumsum(theta)

        for i in range(theta.size):
            A = _affine_matrix(theta_cl[i + 1], x_cl[i], y_cl[i], R_inv=True)
            x_cl[i + 1], y_cl[i + 1] = A @ [-tx[i], -ty[i], 1]

        x_cl, y_cl, theta_cl = x_cl[1:], y_cl[1:], theta_cl[1:]
        return x_cl, y_cl, theta_cl

    def camera2obj(x_ac, y_ac, x_cl, y_cl, theta_cl):
        x_al, y_al = np.empty((2, x_ac.size))

        for i in range(x_ac.size):
            A = _affine_matrix(theta_cl[i], x_cl[i], y_cl[i], R_inv=True)
            x_al[i], y_al[i] = A @ [x_ac[i], y_ac[i], 1]

        return x_al, y_al

    def affine2obj(theta, tx, ty, x_ac, y_ac):
        x_cl, y_cl, theta_cl = affine2camera(theta, tx, ty)
        x_al, y_al = camera2obj(x_ac, y_ac, x_cl, y_cl, theta_cl)
        return x_al, y_al

    theta, tx, ty = reference

    x_al, y_al = affine2obj(theta, tx, ty, traj.r.x, traj.r.y)

    if start_at_origin:
        x_al = x_al - x_al[0]
        y_al = y_al - y_al[0]

    traj.x = x_al
    traj.y = y_al

    moved_traj = Trajectory(
        x=x_al,
        y=y_al,
        ang=traj.ang,
        t=traj.t,
        traj_id=new_traj_id
    )
    return moved_traj



from __future__ import annotations
import json
import csv
import os
from pathlib import Path
from typing import List, NamedTuple, Tuple, Union
import numpy as np

from yupi.vector import Vector
from yupi.exceptions import LoadTrajectoryError
from yupi.features import Features


_threshold = 1e-12


class TrajectoryPoint(NamedTuple):
    """
    Represents a point of a trajectory.

    Parameters
    ----------
    r : Vector
        Positional data.
    ang : Vector
        Angular data.
    v : Vector
        Velocity data.
    t : float
        Time data.
    """

    r: Vector
    ang: Vector
    v: Vector
    t: float


class Trajectory():
    """
    A Trajectory object represents a multidimensional trajectory.
    It can be iterated to obtain the corresponding point for each
    timestep.

    Parameters
    ----------
    x : np.ndarray
        Array containing position data of X axis, by default None
    y : np.ndarray
        Array containing position data of Y axis, by default None.
    z : np.ndarray
        Array containing position data of X axis, by default None.
    points : np.ndarray
        Array containing position data as a list of points, by default
        None
    dimensions : np.ndarray
        Array containing position data as a list of axis, by default
        None
    t : np.ndarray
        Array containing time data, by default None.
    ang : np.ndarray
        Array containing angle data, by default None.
    dt : float
        If no time data is given this represents the time between each
        position data value.
    t0 : float
        If no time data is given this represents the initial time value,
        by default 0.
    traj_id : str
        Id of the trajectory.
    lazy : bool
        Defines if the velocity vector is not recalculated every time
        is asked.

    Attributes
    ----------
    dt : float
        If no time data is given this represents the time between each
        position data value.
    id : str
        Id of the trajectory.

    Examples
    --------
    You can create a trajectory object by giving the arrays that
    represent it:

    >>> x = [0, 1.2, 3, 2.8]
    >>> y = [0, 3.1, 0.7, 1.6]
    >>> Trajectory(x=x, y=y)

    You can also create the trajectory given the points:

    >>> points = [[0, 0], [1.2, 3.1], [3, 0.7], [2.8, 1.6]]
    >>> Trajectory(points=points)

    Or even create it given all the data for each dimension in a single
    source:

    >>> dims = [[0, 1.2, 3, 2.8], [0, 3.1, 0.7, 1.6]]
    >>> Trajectory(dimensions=dims)

    All of these examples create the same trajectory.

    Raises
    ------
    ValueError
        If positional data is given in more than one way.
    ValueError
        If no positional data is given.
    ValueError
        If all the given input data (``x``, ``y``, ``z``, ``t``,
        ``ang``) does not have the same shape.
    ValueError
        If ``t`` and ``dt`` given but ``t`` is not uniformly spaced.
    ValueError
        If ``t`` and ``dt`` given but ``dt`` does not match ``t``
        values delta.
    """

    def __init__(self, x: np.ndarray = None, y: np.ndarray = None,
                 z: np.ndarray = None, points: np.ndarray = None,
                 dimensions: np.ndarray = None, t: np.ndarray = None,
                 ang: np.ndarray = None, dt: float = None, t0: float = 0.0,
                 traj_id: str = None, lazy: bool = False):

        from_xyz = x is not None
        from_points = points is not None
        from_dimensions = dimensions is not None

        if from_xyz + from_points + from_dimensions > 1:
            raise ValueError("Positional data must come only from one way: "
                             "'xyz' data, 'points' data or 'dimensions' data.")

        self.r = None
        data: List[Vector] = [t, ang]
        lengths = [len(item) for item in data if item is not None]

        for i, item in enumerate(data):
            if item is not None:
                data[i] = Vector.create(item, dtype=float)

        if from_xyz:
            dimensions = [d for d in [x, y, z] if d is not None]
            from_dimensions = True

        if from_dimensions and len(dimensions) > 0:
            lengths.extend([len(d) for d in dimensions])
            self.r = Vector.create(dimensions, dtype=float).T

        if from_points:
            lengths.append(len(points))
            self.r = Vector.create(points, dtype=float)

        if self.r is None:
            raise ValueError('No position data were given.')

        if lengths.count(lengths[0]) != len(lengths):
            raise ValueError('All input arrays must have the same shape.')

        self.__t0 = t0
        self.__t = data[0]
        self.ang = data[1]
        self.traj_id= traj_id
        self.lazy = lazy

        if self.__t is None:
            self.dt = dt if dt is not None else 1.0
            self.dt_std = 0
            self.__v: Vector = self.r.delta / self.dt
        else:
            self.dt = np.mean(np.array(self.__t.delta))
            self.dt_std = np.std(np.array(self.__t.delta))
            self.__v: Vector = (self.r.delta.T / self.__t.delta).T

        if t is not None and dt is not None:
            if abs(self.dt - dt) > _threshold:
                raise ValueError("You are giving 'dt' and 't' but 'dt' "
                                "does not match with time values delta.")
            if abs(self.dt_std - 0) > _threshold:
                raise ValueError("You are giving 'dt' and 't' but 't' is "
                                    "not uniformly spaced.")
            if abs(self.__t[0] - t0) > _threshold:
                raise ValueError("You are giving 'dt' and 't' but 't0' is not "
                                 "the same as the first value of 't'.")

        self.features = Features(self)

    @property
    def uniformly_spaced(self) -> bool:
        """bool : True if the time data is uniformly spaced"""
        if self.__t is not None:
            starts_at_zero = self.__t[0] == 0
            std_is_zero = self.dt_std == 0
            if starts_at_zero and std_is_zero:
                return True
            return False
        return True

    def __len__(self):
        return self.r.shape[0]

    def __getitem__(self, index):
        if isinstance(index, int):
            # *dim, *ang, v, t
            data = [self.r[index], [], None, None]

            # Angle
            if self.ang is not None:
                data[1] = self.ang[index]

            # Velocity
            data[2] = self.v[index - 1] if index > 0 else Vector.create([0]*self.dim)

            # Time
            if self.__t is not None:
                data[3] = self.__t[index]
            else:
                data[3] = self.__t0 + self.dt * index

            r, ang, v, t = data
            return TrajectoryPoint(r=r, ang=ang, v=v, t=t)

        if isinstance(index, slice):
            start, stop, step = index.indices(len(self))

            new_points = self.r[start:stop:step]
            new_ang = None
            if self.ang is not None:
                new_ang = self.ang[start:stop:step]
            if self.uniformly_spaced:
                new_dt = self.dt * step
                new_t0 = self.__t0 + start * self.dt
                return Trajectory(points=new_points, ang=new_ang, dt=new_dt,
                                  t0=new_t0)
            new_t = self.t[start:stop:step]
            return Trajectory(points=new_points, ang=new_ang, t=new_t)

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    @property
    def bounds(self) -> List[Tuple[float]]:
        """List[Tuple[float]] : List of tuples indicanting the min and
        max values of each dimension"""
        _bounds = []
        for d in range(self.dim):
            min_bound = min(self.r.component(d))
            max_bound = max(self.r.component(d))
            _bounds.append((min_bound, max_bound))
        return _bounds

    @property
    def dim(self) -> int:
        """int : Trajectory spacial dimensions."""
        return self.r.shape[1]

    @property
    def delta_r(self) -> Vector:
        """Vector: Difference between each couple of consecutive points
        in the Trajectory."""
        return self.r.delta

    @property
    def delta_v(self) -> Vector:
        """Vector: Difference between each couple of consecutive sample
        in the velocity vector of the Trajectory."""
        return self.v.delta

    @property
    def delta_ang(self) -> Union[Vector, None]:
        """Union[Vector, None] : Difference between each couple of
        consecutive samples in the ``ang`` vector of the Trajectory."""
        if self.ang is not None:
            return self.ang.delta
        return None

    def recalculate_velocity(self) -> Vector:
        """
        Recalculates the velocity according time data or `dt` if time
        data is not available.

        Returns
        -------
        Vector
            Velocity vector.
        """

        if self.uniformly_spaced:
            self.__v: Vector = self.r.delta / self.dt
        else:
            self.__v: Vector = (self.r.delta.T / self.__t.delta).T
        return self.__v

    @property
    def v(self) -> Vector:
        """Vector : Velocity vector"""
        if self.lazy:
            return self.__v
        return self.recalculate_velocity()

    @property
    def t(self) -> Vector:
        """Vector : Time vector"""
        if self.__t is None:
            dt_vec = [self.__t0 + self.dt*i for i in range(len(self))]
            self.__t = Vector.create(dt_vec)
        return self.__t

    @property
    def ang_velocity(self) -> Union[Vector, None]:
        """Union[Vector, None] : Computes the angular velocity from the
        ``ang`` vector of the Trajectory."""
        if self.ang is not None:
            return self.ang.delta / self.dt
        return None

    def add_polar_offset(self, radius: float, angle: float) -> None:
        """
        Adds an offset given a point in polar coordinates.

        Parameters
        ----------
        radius : float
            Point's radius.
        angle : float
            Point's angle.

        Raises
        ------
        TypeError
            If the trajectory is not 2 dimensional.
        """

        if self.dim != 2:
            raise TypeError('Polar offsets can only be applied on 2 '
                            'dimensional trajectories')

        # From cartesian to polar
        x, y = self.r.x, self.r.y
        rad = np.hypot(x, y)
        ang = np.arctan2(y, x)

        rad += radius
        ang += angle

        # From polar to cartesian
        x = rad * np.cos(ang)
        y = rad * np.sin(ang)
        self.r = Vector.create([x,y]).T

    def rotate(self, angle: float):
        """
        Rotates the trajectory around the center coordinates [0,0]

        Parameters
        ----------
        angle : float
            Angle in radians to rotate the trajectory.
        """
        self.add_polar_offset(0, angle)

    def copy(self) -> Trajectory:
        """
        Returns a copy of the trajectory.

        Returns
        -------
        Trajectory
            Copy of the trajectory.
        """

        return Trajectory(points=self.r, t=self.__t, ang=self.ang, dt=self.dt,
                          lazy=self.lazy)

    def _operable_with(self, other: Trajectory, threshold=None) -> bool:
        if threshold is None:
            threshold = _threshold

        if self.r.shape != other.r.shape:
            return False

        self_time = self.t
        other_time = other.t

        diff = np.abs(self_time - other_time)
        return all(diff < threshold)

    def __iadd__(self, other):
        if isinstance(other, (int, float)):
            self.r += other
            return self

        if isinstance(other, (list, tuple, np.ndarray)):
            offset = np.array(other, dtype=float)
            if len(offset) != self.dim:
                raise ValueError('Offset must be the same shape as the other '
                                 'trajectory points')
            self.r += offset
            return self

        if isinstance(other, Trajectory):
            if not self._operable_with(other):
                raise ValueError('Incompatible trajectories')
            self.r += other.r
            return self

        raise TypeError("unsoported operation (+) between 'Trajectory' and "
                        f"'{type(other).__name__}'")

    def __isub__(self, other):
        if isinstance(other, (int, float)):
            self.r -= other
            return self

        if isinstance(other, (list, tuple, np.ndarray)):
            offset = np.array(other, dtype=float)
            if len(offset) != self.dim:
                raise ValueError('Offset must be the same shape as the other '
                                 'trajectory points')
            self.r -= offset
            return self

        if isinstance(other, Trajectory):
            if not self._operable_with(other):
                raise ValueError('Incompatible trajectories')
            self.r -= other.r
            return self

        raise TypeError("unsoported operation (-) between 'Trajectory' and "
                        f"'{type(other).__name__}'")

    def __add__(self, other):
        traj = self.copy()
        traj += other
        return traj

    def __sub__(self, other):
        traj = self.copy()
        traj -= other
        return traj

    def __radd__(self, other):
        return self + other

    def __rsub__(self, other):
        return self - other

    def __imul__(self, other):
        if isinstance(other, (int, float)):
            self.r *= other
            return self

        raise TypeError("unsoported operation (*) between 'Trajectory' and "
                        f"'{type(other).__name__}'")

    def __mul__(self, other):
        traj = self.copy()
        traj *= other
        return traj

    def __rmul__(self, other):
        return self * other

    def turning_angles(self, accumulate=False, degrees=False,
                       centered=False, wrap=True):
        """
        Return the sequence of turning angles that forms the trajectory.

        Parameters
        ----------
        traj : Trajectory
            Input trajectory.
        accumulate : bool, optional
            If True, turning angles are measured with respect to an axis
            defined by the initial velocity (i.e., angles between initial
            and current velocity). Otherwise, relative turning angles
            are computed (i.e., angles between succesive velocity
            vectors). By default False.
        degrees : bool, optional
            If True, angles are given in degrees. Otherwise, the units
            are radians. By default False.
        centered : bool, optional
            If True, angles are wrapped on the interval ``[-pi, pi]``.
            Otherwise, the interval ``[0, 2*pi]`` is chosen. By default
            False.
        wrap : bool, optional
            If True, angles are wrapped in a certain interval (depending
            on ``centered`` param). By default True.

        Returns
        -------
        np.ndarray
            Turning angles where each position in the array correspond
            to a given time instant.
        """

        dx = self.delta_r.x
        dy = self.delta_r.y
        theta = np.arctan2(dy, dx)

        if not accumulate:
            theta = np.ediff1d(theta)  # Relative turning angles
        else:
            theta -= theta[0]          # Accumulative turning angles

        if degrees:
            theta = np.rad2deg(theta)

        if not wrap:
            return theta

        discont = 360 if degrees else 2 * np.pi
        if not centered:
            return theta % discont

        discont_half = discont / 2
        return -((discont_half - theta) % discont - discont_half)

    def _save_json(self, path: str):
        def convert_to_list(vec: Vector):
            if vec is None:
                return vec

            if len(vec.shape) == 1:
                return list(vec)
            return {d: list(v) for d, v in enumerate(vec)}

        ang = None if self.ang is None else self.ang.T
        json_dict = {
            'id': self.traj_id,
            'dt': self.dt,
            'r': convert_to_list(self.r.T),
            'ang': convert_to_list(ang),
            't': convert_to_list(self.__t)
        }
        with open(str(path), 'w') as traj_file:
            json.dump(json_dict, traj_file)

    def _save_csv(self, path):
        with open(str(path), 'w', newline='') as traj_file:
            writer = csv.writer(traj_file, delimiter=',')
            ang_shape = 0 if self.ang is None else self.ang.shape[1]
            writer.writerow([self.traj_id, self.dt, self.dim, ang_shape])
            for tp in self:
                row = np.hstack(np.array([tp.r, tp.ang, tp.t]))
                writer.writerow(row)

    def save(self, file_name: str, path: str = '.', file_type: str = 'json',
             overwrite: bool = True):
        """
        Saves the trajectory to disk.

        Parameters
        ----------
        file_name : str
            Name of the file.
        path : str
            Path where to save the trajectory, by default ``'.'``.
        file_time : str
            Type of the file, by default ``json``.

            The only types avaliable are: ``json`` and ``csv``.
        overwrite : bool
            Wheter or not to overwrite the file if it already exists,
            by default True.

        Raises
        ------
        ValueError
            If ``override`` parameter is ``False`` and the file already
            exists.
        ValueError
            If ``file_type`` is not ``json`` or ``csv``.

        Examples
        --------
        >>> t = Trajectory(x=[0.37, 1.24, 1.5])
        >>> t.save('my_track')
        """

        # Build full path
        full_path = Path(path) / Path(f'{file_name}.{file_type}')

        # Check file existance
        if not overwrite and full_path.exists():
            raise ValueError(f"File '{str(full_path)}' already exist")

        if file_type == 'json':
            self._save_json(full_path)
        elif file_type == 'csv':
            self._save_csv(full_path)
        else:
            raise ValueError(f"Invalid export file type '{file_type}'")

    @staticmethod
    def save_trajectories(trajectories: list, folder_path: str = '.',
                          file_type: str = 'json', overwrite: bool = True):
        """
        Saves a list of trajectories to disk. Each Trajectory object
        will be saved in a separate file inside the given folder.

        Parameters
        ----------
        trajectories : list[Trajectory]
            List of Trajectory objects that will be saved.
        folder_path : str
            Path where to save all the trajectory, by default ``'.'``.
        file_type : str
            Type of the file, by default ``json``.

            The only types avaliable are: ``json`` and ``csv``.
        overwrite : bool
            Wheter or not to overwrite the file if it already exists,
            by default True.

        Examples
        --------
        >>> t1 = Trajectory(x=[0.37, 1.24, 1.5])
        >>> t2 = Trajectory(x=[1, 2], y=[3, 4])
        >>> Trajectory.save_trajectories([t1, t2])
        """

        for i, traj in enumerate(trajectories):
            path = str(Path(folder_path))
            name = str(Path(f'trajectory_{i}'))
            traj.save(name, path, file_type, overwrite)

    @staticmethod
    def _load_json(path: str):
        with open(path, 'r') as traj_file:
            data = json.load(traj_file)

            traj_id = data['id']
            dt = data['dt']
            t = data['t']
            ang = None

            if data['ang'] is not None:
                ang_values = list(data['ang'].values())
                ang = Vector.create(ang_values).T

            dims = list(data['r'].values())

            return Trajectory(dimensions=dims, t=t, ang=ang, dt=dt,
                              traj_id=traj_id)

    @staticmethod
    def _load_csv(path: str):
        with open(path, 'r') as traj_file:

            def check_empty_val(val, cast=True):
                if val == '':
                    return None
                return float(val) if cast else val

            r, ang, t = [], [], []
            traj_id, dt, dim = None, None, None

            for i, row in enumerate(csv.reader(traj_file)):
                if i == 0:
                    traj_id = check_empty_val(row[0], cast=False)
                    dt = check_empty_val(row[1])
                    dim = int(row[2])
                    ang_dim = int(row[3])
                    r = [[] for _ in range(dim)]
                    ang = [[] for _ in range(ang_dim)]
                    continue

                for j in range(dim):
                    r[j].append(float(row[j]))

                for j, k in enumerate(range(dim, dim + ang_dim)):
                    ang[j] = row[k]

                t.append(float(row[-1]))

            if not ang:
                ang = None

            return Trajectory(dimensions=r, t=t, ang=ang, dt=dt,
                              traj_id=traj_id)

    @staticmethod
    def load(file_path: str):
        """
        Loads a trajectory

        Parameters
        ----------
        file_path : str
            Path of the trajectory file

        Returns
        -------
        Trajectory
            Loaded Trajectory object.

        Raises
        ------
        ValueError
            If ``file_path`` is a non existing path.
        ValueError
            If ``file_path`` is a not a file.
        ValueError
            If ``file_path`` extension is not ``json`` or ```csv``.
        """

        path = Path(file_path)

        # Check valid path
        if not path.exists():
            raise ValueError('Path does not exist.')
        if not path.is_file():
            raise ValueError("Path must be a file.")

        file_type = path.suffix

        try:
            if file_type == '.json':
                return Trajectory._load_json(file_path)
            elif file_type == '.csv':
                return Trajectory._load_csv(file_path)
            else:
                raise ValueError("Invalid file type.")
        except (json.JSONDecodeError, KeyError, ValueError, IndexError) as exc:
            raise LoadTrajectoryError(path) from exc

    @staticmethod
    def load_folder(folder_path='.', recursively: bool = False):
        """
        Loads all the trajectories from a folder.

        Parameters
        ----------
        folder_path : str
            Path of the trajectories folder.
        recursively : bool
            If True then subfolders are analized recursively, by
            default False.

        Returns
        -------
        List[Trajectory]
            List of the loaded trajectories.
        """

        trajectories = []
        for root, _, files in os.walk(folder_path):
            for file in files:
                path = str(Path(root) / Path(file))
                try:
                    trajectories.append(Trajectory.load(path))
                except LoadTrajectoryError as load_exception:
                    print(f'Ignoring: {load_exception.path}')
            if not recursively:
                break
        return trajectories



import abc
from typing import Callable, Tuple
import numpy as np
from yupi import Trajectory



class Generator(metaclass=abc.ABCMeta):
    """
    Abstract class to model a Trajectory Generator. Classes inheriting
    from this class should implement ``generate`` method.

    Parameters
    ----------
    T : float
        Total duration of each Trajectory.
    dim : int, optional
        Dimension of each Trajectory, by default 1.
    N : int, optional
        Number of trajectories, by default 1.
    dt : float, optional
        Time step of the Trajectory, by default 1.0.

    Attributes
    ----------
    T : float
        Total duration of each Trajectory.
    dim : int, optional
        Dimension of each Trajectory, by default 1.
    N : int, optional
        Number of trajectories, by default 1.
    dt : float, optional
        Time step of the Trajectory, by default 1.0.
    n  : int
        Number of samples on each Trajectory.
    """

    def __init__(self, T: float, dim: int = 1, N: int = 1, dt: float = 1.0):
        # Simulation parameters
        self.T = T            # Total time
        self.dim = dim        # Trajectory dimension
        self.N = N            # Number of trajectories
        self.dt = dt          # Time step of the simulation
        self.n = int(T / dt)  # Number of time steps

    @abc.abstractmethod
    def generate(self):
        """
        Abstract method that is implemented on inheriting classes.
        It should compute a list of ``N`` Trajectory objects with the
        given parameters using a method specific to the inheriting class.
        """


class RandomWalkGenerator(Generator):
    """
    Multidimensional Random Walk Generator.

    Parameters
    ----------
    T : float
        Total duration of each Trajectory.
    dim : int, optional
        Dimension of each Trajectory, by default 1.
    N : int, optional
        Number of trajectories, by default 1.
    dt : float, optional
        Time step of the Trajectory, by default 1.0.
    actions_prob : np.ndarray, optional
        Probability of each action (i.e., decrease, stead or increase)
        to be taken, according to every axis. If this parameter is not
        passed the walker will assume uniform probability for each
        action, by default None.
    step_length_func : Callable[[Tuple], np.ndarray], optional
        Function that returns the distribution of step lengths that
        will be taken by the walker on each time step, dimension and
        instance of a trajectory. Expected shape of the return value is
        (int(T/dt)-1, dim, N), by default np.ones.
    step_length_kwargs : dict, optional
        Key-word arguments of the ``step_length_func``, by default
        ``{}``.
    """

    def __init__(self, T: float, dim: int = 1, N: int = 1, dt: float = 1,
                 actions_prob: np.ndarray = None,
                 step_length_func: Callable[[Tuple], np.ndarray] = np.ones,
                 **step_length_kwargs):

        super().__init__(T, dim, N, dt)

        # Main id of generated trajectories
        self.traj_id = 'RandomWalk'

        # Dynamic variables
        self.t = np.arange(self.n) * dt      # Time array
        self.r = np.zeros((self.n, dim, N))  # Position array

        # Model parameters
        actions = np.array([-1,0,1])

        if actions_prob is None:
            actions_prob = np.tile([1/3, 1/3, 1/3], (dim, 1))

        actions_prob = np.asarray(actions_prob, dtype=np.float32)

        if actions_prob.shape[0] != dim:
            raise ValueError('actions_prob must have shape like (dims, 3)')
        if actions_prob.shape[1] != actions.shape[0]:
            raise ValueError('actions_prob must have shape like (dims, 3)')

        shape_tuple = (self.n - 1, dim, N)
        step_length = step_length_func(shape_tuple, **step_length_kwargs)

        self.actions = actions
        self.actions_prob = actions_prob
        self.step_length = step_length


    # Compute vector position as a function of time for
    # All the walkers of the ensemble
    def _get_r(self):
        # Get movements for every space coordinates according
        # To the sample space of probabilities in self.actions_prob
        dr = [np.random.choice(self.actions, p=p, size=(self.n - 1, self.N))
              for p in self.actions_prob]

        # Set time/coordinates as the first/second axis
        dr = np.swapaxes(dr, 0, 1)

        # Scale displacements according to the jump length statistics
        dr = dr * self.step_length

        # Integrate displacements to get position vectors
        self.r[1:] = np.cumsum(dr, axis=0)
        return self.r


    # Get position vectors and generate RandomWalk object
    def generate(self):
        # Get position vectors
        r = self._get_r()

        # Generate RandomWalk object
        trajs = []
        for i in range(self.N):
            points = r[:, :, i]
            trajs.append(Trajectory(points=points, dt=self.dt, t=self.t,
                                    traj_id=f"{self.traj_id} {i + 1}"))
        return trajs


class _LangevinGenerator(Generator):

    def __init__(self, 
                 T: float, 
                 dim: int = 1, 
                 N: int = 1, 
                 dt: float = 1., 
                 tau: float = 1., 
                 noise_scale: float = 1., 
                 v0: np.ndarray = None, 
                 r0: np.ndarray = None):

        super().__init__(T, dim, N, dt)

        # Main id of generated trajectories
        self.traj_id = 'Langevin'

        # Model parameters
        self.tau = tau                  # Relaxation time
        self.noise_scale = noise_scale  # Noise scale parameter

        # Intrinsic reference parameters
        self.t_scale = tau                                  # Time scale
        self.v_scale = noise_scale * np.sqrt(self.t_scale)  # Speed scale
        self.r_scale = self.v_scale * self.t_scale          # Length scale

        # Simulation parameters
        self.dt = dt / self.t_scale    # Dimensionless time step
        self.shape = (self.n, dim, N)  # Shape of dynamic variables

        # Dynamic variables
        self.t = np.arange(self.n) * self.dt  # Time array
        self.r = np.empty(self.shape)         # Position array
        self.v = np.empty(self.shape)         # Velocity array
        self.noise = None                     # Noise array (filled in _set_noise method)

        # Initial conditions
        self.r0 = r0           # Initial position
        self.v0 = v0           # Initial velocity
        self._set_init_cond()  # Check and set initial conditions


    # Set initial conditions
    def _set_init_cond(self):
        if self.r0 is None:
            self.r[0] = np.zeros((self.dim, self.N))  # Default initial positions
        elif np.shape(self.r0) == (self.dim, self.N) or np.shape(self.r0) == ():
            self.r[0] = self.r0                       # User initial positions
        else:
            raise ValueError('r0 is expected to be a float or an '
                            f'array of shape {(self.dim, self.N)}.')
        
        if self.v0 is None:
            self.v[0] = np.random.normal(size=(self.dim, self.N))  # Default initial velocities
        elif np.shape(self.v0) == (self.dim, self.N) or np.shape(self.v0) == ():
            self.v[0] = self.v0                                    # User initial velocities
        else:
            raise ValueError('v0 is expected to be a float or an '
                            f'array of shape {(self.dim, self.N)}.')


    # Fill noise array with custom noise properties
    def _set_noise(self):
        self.noise = np.random.normal(size=self.shape)


    # Solve dimensionless Langevin Equation using 
    # the numerical method of Euler-Maruyama
    def _solve(self):
        for i in range(self.n - 1):
            # Solving for position
            self.r[i + 1] = self.r[i] + \
                            self.v[i] * self.dt

            # Solving for velocity
            self.v[i + 1] = self.v[i] + \
                            -self.v[i] * self.dt + \
                            np.sqrt(self.dt) * self.noise[i]


    # Scale by intrinsic reference quantities
    def _set_scale(self):
        self.r *= self.r_scale
        self.v *= self.v_scale
        self.t *= self.t_scale
        self.dt *= self.t_scale


    # Simulate the process
    def _simulate(self):
        self._set_noise()  # Set the attribute self.noise
        self._solve()      # Solve the Langevin equation
        self._set_scale()  # Scaling


    # Generate yupi Trajectory objects
    def generate(self):
        self._simulate()

        trajs = []
        for i in range(self.N):
            points = self.r[:, :, i]
            trajs.append(Trajectory(points=points, dt=self.dt,
                                    traj_id=f"{self.traj_id} {i + 1}"))
        return trajs


class LangevinGenerator(_LangevinGenerator):
    """
    Random Walk class from a multidimensional Langevin Equation. 
    Boundary conditions to model confined or semi-infinite processes 
    are supported.

    Parameters
    ----------
    T : float
        Total duration of trajectories.
    dim : int, optional
        Trajectories dimension, by default 1.
    N : int, optional
        Number of simulated trajectories, by default 1.
    dt : float, optional
        Time step, by default 1.0.
    tau : float, optional
        Persistence time, by default 1.
    noise_scale : float, optional
        Noise intensity (i.e., scale parameter of noise pdf), by default 1.
    bounds: np.ndarray, optional
        Lower and upper reflecting boundaries that confine the trajectories. If None 
        is passed, trajectories are simulated in a free space. By default None.
    bounds_extent: np.ndarray, optional
        Decay length of boundary forces, by default None.
    bounds_strength: np.ndarray, optional
        Boundaries strength, by default None.
    v0 : np.ndarray, optional
        Initial velocities, by default None.
    r0 : np.ndarray, optional
        Initial positions, by default None.
    """

    def __init__(self, 
                 T: float, 
                 dim: int = 1, 
                 N: int = 1, 
                 dt: float = 1., 
                 tau: float = 1., 
                 noise_scale: float = 1.,  
                 bounds: np.ndarray = None, 
                 bounds_extent: np.ndarray = None, 
                 bounds_strength: np.ndarray = None, 
                 v0: np.ndarray = None, 
                 r0: np.ndarray = None):

        super().__init__(T, dim, N, dt, tau, noise_scale, v0, r0)

        # Verify if there is any boundary
        self.bounds = np.float32(bounds)                     # Convert None into np.nan
        self.has_bounds = not np.all(np.isnan(self.bounds))  # Check for all bounds

        if self.has_bounds:
            # Broadcast and scale bounds properties
            ones = np.ones((2, self.dim))
            self.bounds = self.bounds * ones / self.r_scale
            self.bounds_ext = np.float32(bounds_extent) * ones / self.r_scale
            self.bounds_stg = np.float32(bounds_strength) * ones * (self.t_scale/self.v_scale)
            
            # Check is initial positions are within bounds
            self._check_r0()


    # Check if all initial positions are inside boundaries
    def _check_r0(self):
        # Unpack lower and upper bounds
        lb, ub = self.bounds

        # Find axes without boundaries
        idx_lb = np.where(np.isnan(lb))
        idx_ub = np.where(np.isnan(ub))

        # Ignore position components when no boundaries are specified
        r_lb = np.delete(self.r[0], idx_lb, axis=0)
        r_ub = np.delete(self.r[0], idx_ub, axis=0)

        # Same for bounds
        lb = np.delete(lb, idx_lb)
        ub = np.delete(ub, idx_ub)

        # Check if all positions are within both type of boundaries
        is_above_lb = np.all(lb[:,None] <= r_lb)
        is_bellow_ub = np.all(ub[:,None] >= r_ub)

        if not is_above_lb:
            raise ValueError('Initial positions must be above lower bounds.')
        
        if not is_bellow_ub:
            raise ValueError('Initial positions must be bellow upper bounds.')


    # Get net force from the boundaries
    def _bound_force(self, r, tolerance=10):
        # Return zero force if there is no bounds
        if not self.has_bounds:
            return 0.

        # Set r to have shape = (N, dim)
        r = r.T

        # Lower and upper bound limits, extents and strengths
        lb, ub = self.bounds
        ext_lb, ext_ub = self.bounds_ext
        stg_lb, stg_ub = self.bounds_stg

        # Get distance from the bounds and scale
        # by the bound extent parameter
        dr_lb = (r - lb) / ext_lb
        dr_ub = (r - ub) / ext_ub

        # An exponential models the force from the wall. 
        # Get zero force if there is no bound or the particle 
        # is far enough.
        force_lb = np.where(
                        np.isnan(lb) | (dr_lb > tolerance), 
                        0., 
                        stg_lb * np.exp(-dr_lb))

        force_ub = np.where(
                        np.isnan(ub) | (-dr_ub > tolerance), 
                        0., 
                        -stg_ub * np.exp(dr_ub))

        # Adding boundary effects and transpose to recover 
        # shape as (dim, N)
        bound_force = (force_lb + force_ub).T
        return bound_force


    # Solve dimensionless Langevin Equation using 
    # the numerical method of Euler-Maruyama
    def _solve(self):
        for i in range(self.n - 1):
            # Solving for position
            self.r[i + 1] = self.r[i] + \
                            self.v[i] * self.dt

            # Solving for velocity
            self.v[i + 1] = self.v[i] + \
                            -self.v[i] * self.dt + \
                            np.sqrt(self.dt) * self.noise[i] + \
                            self._bound_force(self.r[i]) * self.dt


class DiffDiffGenerator(Generator):
    """
        Random Walk class for the Diffusing Diffusivity model.

    Parameters
    ----------
    T : float
        Total duration of each Trajectory.
    dim : int, optional
        Dimension of each Trajectory, by default 1.
    N : int, optional
        Number of trajectories, by default 1.
    dt : float, optional
        Time step of the Trajectory, by default 1.0.
    tau : float, optional
        Relaxation characteristic time of the auxiliary variable, by default 1.
    noise_scale : float, optional
        Scale parameter of the auxiliary variable noise, by default 1.
    dim_aux: int, optional
        Dimension of the auxiliary process, which is the square of the diffusivity, by default 1.
    r0 : np.ndarray, optional
        Initial positions, by default None.
    """

    def __init__(self, 
                 T: float, 
                 dim: int = 1, 
                 N: int = 1, 
                 dt: float = 1., 
                 tau: float = 1., 
                 noise_scale: float = 1., 
                 dim_aux: int = 1, 
                 r0: np.ndarray = None):

        super().__init__(T, dim, N, dt)

        # Model parameters
        self.tau = tau                  # Relaxation time
        self.noise_scale = noise_scale  # Noise scale parameter of auxiliary variable

        # Intrinsic reference parameters
        self.t_scale = tau                         # Time scale
        self.r_scale = noise_scale * self.t_scale  # Length scale

        # Simulation parameters
        self.dt = dt / self.t_scale    # Dimensionless time step
        self.shape = (self.n, dim, N)  # Shape of dynamic variables
        self.dim_aux = dim_aux         # Dimension of the aux variable

        # Dynamic variables
        self.t = np.arange(self.n, dtype=np.float32)  # Time array
        self.r = np.empty(self.shape)                 # Position array
        self.Y = np.empty((dim_aux, N))               # Aux variable: square of diffusivity
        self.noise_r = None                           # Noise array for position (filled in _set_noise method)
        self.noise_Y = None                           # Noise array for aux variable (filled in _set_noise method)

        # Initial conditions
        self.Y = np.random.normal(size=(dim_aux, N))  # Initial aux variable configuration
        self.D = np.sum(self.Y**2, axis=0)            # Initial diffusivity configuration

        if r0 is None:
            self.r[0] = np.zeros((dim, N))                    # Default initial positions
        elif np.shape(r0) == (dim, N) or np.shape(r0) == ():
            self.r[0] = r0                                    # User initial positions
        else:
            raise ValueError(f'r0 is expected to be a float or an array of shape {(self.dim, self.N)}.')


    # Fill noise arrays
    def _set_noise(self):
        dist = np.random.normal
        self.noise_r = dist(size=self.shape)
        self.noise_Y = dist(size=(self.n, self.dim_aux, self.N))
    

    # Solve coupled Langevin equations
    def _solve(self):
        for i in range(self.n - 1):
            # Solving for position
            self.r[i + 1] = self.r[i] + \
                            np.sqrt(2 * self.D * self.dt) * self.noise_r[i]

            # Solving for auxliliary variable
            self.Y = self.Y + \
                     -self.Y * self.dt + \
                     np.sqrt(self.dt) * self.noise_Y[i]
            
            # Updating the diffusivities
            self.D = np.sum(self.Y**2, axis=0)


    # Scale by intrinsic reference quantities
    def _set_scale(self):
        self.r *= self.r_scale
        self.t *= self.t_scale
        self.dt *= self.t_scale


    # Simulate the process
    def _simulate(self):
        self._set_noise()  # Set the attribute self.noise
        self._solve()      # Solve the Langevin equation
        self._set_scale()  # Scaling


    # Generate yupi Trajectory objects
    def generate(self):
        self._simulate()

        trajs = []
        for i in range(self.N):
            points = self.r[:, :, i]
            trajs.append(Trajectory(points=points, dt=self.dt,
                                    traj_id=f"DiffDiff {i + 1}"))
        return trajs


from typing import List
import numpy as np
from yupi.trajectory import Trajectory, _threshold

def _check_uniform_time_spaced(func):
    def wrapper(trajs: List[Trajectory], *args, **kwargs):
        if any((abs(t.dt_std - 0) > _threshold for t in trajs)):
            raise ValueError('All trajectories must be uniformly time spaced')
        return func(trajs, *args, **kwargs)
    wrapper.__doc__ = func.__doc__
    return wrapper

def _check_same_dt(func):
    def wrapper(trajs: List[Trajectory], *args, **kwargs):
        if trajs:
            dt = trajs[0].dt
            if any((abs(t.dt - dt) > _threshold for t in trajs)):
                raise ValueError("All trajectories must have the same 'dt'")
        return func(trajs, *args, **kwargs)
    wrapper.__doc__ = func.__doc__
    return wrapper

def _check_same_dim(func):
    def wrapper(trajs: List[Trajectory], *args, **kwargs):
        if trajs:
            dim = trajs[0].dim
            if any((t.dim != dim for t in trajs)):
                raise ValueError("All trajectories must have the same dimensions")
        return func(trajs, *args, **kwargs)
    wrapper.__doc__ = func.__doc__
    return wrapper

def _check_exact_dim(dim):
    def _check_exact_dim_decorator(func):
        def wrapper(trajs: List[Trajectory], *args, dim=dim, **kwargs):
            if any((t.dim != dim for t in trajs)):
                raise ValueError(f"All trajectories must be {dim}-dimensional")
            return func(trajs, *args, **kwargs)
        wrapper.__doc__ = func.__doc__
        return wrapper
    return _check_exact_dim_decorator

def _check_same_r0(func):
    def wrapper(trajs: List[Trajectory], *args, **kwargs):
        if trajs:
            r0 = trajs[0].r[0]
            if any((abs(t.r[0] - r0) > _threshold for t in trajs)):
                raise ValueError("All trajectories must have the same initial position")
        return func(trajs, *args, **kwargs)
    wrapper.__doc__ = func.__doc__
    return wrapper

def _check_same_lenght(func):
    def wrapper(trajs: List[Trajectory], *args, **kwargs):
        if trajs:
            length = len(trajs[0])
            if any((abs(len(t) - length) > _threshold for t in trajs)):
                raise ValueError("All trajectories must have the same length")
        return func(trajs, *args, **kwargs)
    wrapper.__doc__ = func.__doc__
    return wrapper

def _check_same_t(func):
    def wrapper(trajs: List[Trajectory], *args, **kwargs):
        if trajs:
            zipped_trajs = zip(trajs[:-1], trajs[1:])
            if not all((np.allclose(t0.t, t1.t, atol=_threshold) for t0, t1 in zipped_trajs)):
                raise ValueError("All trajectories must have the same 't' values")
        return func(trajs, *args, **kwargs)
    wrapper.__doc__ = func.__doc__
    return wrapper



from yupi import Trajectory

def subsample(traj: Trajectory, step=1, new_traj_id: str = None):
    """
    Sample the trajectory ``traj`` by removing evenly spaced
    points according to ``step``.

    Parameters
    ----------
    traj : Trajectory
        Input trajectory.
    step : int, optional
        Number of sample points or period. By default 1.
    Returns
    -------
    Trajectory
        Output trajectory.
    """

    points = traj.r[::step]
    ang = traj.ang[::step] if traj.ang is not None else None
    t = traj.t[::step] if traj.t is not None else None

    subsampled_traj = Trajectory(
        points=points,
        t=t,
        ang=ang,
        dt=step*traj.dt,
        traj_id=new_traj_id
    )
    return subsampled_traj



import matplotlib.pyplot as plt

LINE = '-'
DOTTED = 'o'
LINE_DOTTED = '-o'

# Main colors
BLUE = '#3489b1'
MAGENTA = '#a14383'
YELLOW = '#e5d935'
RED = '#af3c3c'
GREEN = '#54ac43'
ORANGE = '#e88d26'
MID_BLUE = '#3c4baf'

#  Light colors
LIGHT_BLUE = '#99d2ec'
LIGHT_MAGENTA = '#eaa0d2'
LIGHT_YELLOW = '#fdf584'
LIGHT_RED = '#ea8080'
LIGHT_GREEN = '#a6ec98'
LIGHT_ORANGE = '#f7c790'

YUPI_COLORS = [
    BLUE,
    MAGENTA,
    YELLOW,
    RED,
    GREEN,
    ORANGE,
    MID_BLUE
]

YUPI_LIGHT_COLORS = [
    LIGHT_BLUE,
    LIGHT_MAGENTA,
    LIGHT_YELLOW,
    LIGHT_RED,
    LIGHT_GREEN,
    LIGHT_ORANGE
]

def _plot_basic_properties(func):
    def wrapper(*args, title: str = '', xlabel: str = '', ylabel: str = '',
                grid: bool = False, show: bool = True, legend: bool = False,
                xscale: str = None, yscale: str = None, xlim: tuple = None,
                ylim: tuple = None, **kwargs):
        func(*args, **kwargs)
        plt.grid(grid)
        plt.title(title)
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        if legend:
            plt.legend()
        if xlim is not None:
            plt.xlim(xlim)
        if ylim is not None:
            plt.ylim(ylim)
        if xscale is not None:
            plt.xscale(xscale)
        if yscale is not None:
            plt.yscale(yscale)
        if show:
            plt.show()

    wrapper.__doc__ = func.__doc__
    return wrapper



import abc
import logging
import cv2
import numpy as np


def _resize_frame(frame, scale=1):
    h, w = frame.shape[:2]
    w_, h_ = int(scale * w), int(scale * h)
    short_frame = cv2.resize(frame, (w_, h_), interpolation=cv2.INTER_AREA)
    return short_frame


def _change_colorspace(image, color_space):
    if color_space == 'BGR':
        return image
    if color_space == 'GRAY':
        return cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    if color_space == 'HSV':
        return cv2.cvtColor(image, cv2.COLOR_BGR2HSV)


class BackgroundEstimator():
    """
    This class provides static methods to determine the background in image
    sequences. It estimates the temporal median of the sequence.
    """

    def __init__(self):
        pass

    @staticmethod
    def from_video(video_path, samples, start_in=0):
        """
        This method takes a video indicated by ``video_path`` and
        uniformely take a number of image samples according to the
        parameter ``samples``. Then, it computes the temporal median of
        the images in order to determine de background.

        Parameters
        ----------
        video_path : str
            Path to the video file
        samples : int
            Number of samples to get from the video.
        start_in : int, optional
            If passed, the method will start sampling after the frame
            indicated by this value, by default 0.
        """

        # Create a cv2 Video Capture Object
        cap = cv2.VideoCapture(video_path)

        total_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
        effective_frames = total_frames - start_in
        spacing = effective_frames / samples

         # Store frames in a list
        frames = []
        for i in range(samples):
            cap.set(cv2.CAP_PROP_POS_FRAMES, i * spacing + start_in)
            _, frame = cap.read()
            frames.append(frame)

        # Calculate the median along time
        return np.median(frames, axis=0).astype(dtype=np.uint8)

class TrackingAlgorithm(metaclass=abc.ABCMeta):
    """
    Abstract class to model a Tracking Algorithm. Classes inheriting
    from this class should implement ``detect`` method.
    """

    def __init__(self):
        pass

    def get_centroid(self, bin_img):
        """
        Computes the centroid of a binary image using ``cv2.moments``.

        Parameters
        ----------
        bin_img : np.ndarray
            Binary image used to compute a centroid

        Returns
        -------
        tuple
            x, y coordinates of the centroid
        """

        # Calculate moments
        M = cv2.moments(bin_img)

        # Check if something was over the threshold
        if M['m00'] != 0:
            # Calculate x,y coordinate of center
            cX = int(M['m10'] / M['m00'])
            cY = int(M['m01'] / M['m00'])
            return cX, cY
        logging.warning("Nothing was over threshold. Algorithm: %s",
                        type(self).__name__)
        return None

    def preprocess(self, frame, roi_bound, preprocessing):
        frame = frame.copy()
        if roi_bound:
            xmin, xmax, ymin, ymax = roi_bound
            frame = frame[ymin:ymax, xmin:xmax, :]
        if preprocessing is not None:
            frame = preprocessing(frame)
        return frame

    @abc.abstractmethod
    def detect(self, frame, roi_bound=None, preprocessing=None):
        """
        Abstract method that is implemented on inheriting classes.
        It should compute the location (in the image ``frame``)
        of the object being tracked.

        Parameters
        ----------
        frame : np.ndarray
            Image where the algorithm must identify the object
        roi_bound : tuple, optional
            Coordinates of the region of interest of the frame. The
            expected format if a tuple with the form (xmin, xmax, ymin,
            ymax). If passed the algorithm will crop this region of the
            frame and will proceed only in this region, providing the
            estimations refered to this region instead of the whole
            image, by default None.
        preprocessing : func
            A function to be applied to the frame (Or cropped version
            of it if roi_bound is passed) before detecting the object
            on it, by default None.
        """


class ColorMatching(TrackingAlgorithm):
    """
    Identifies the position of an object by thresholding pixel
    color values in the pre-defined ranges.

    Parameters
    ----------
    lower_bound : tuple, optional
        Minimum value of pixel color to be considered as part of
        the object, by default (0,0,0)
    upper_bound : tuple, optional
        Maximum value of pixel color to be considered as part of
        the object, by default (255,255,255)
    color_space : str, optional
        Color space to be used before thresholding with the given
        bounds. The image will be automatically converted to this
        color space, by default 'BGR'.
    max_pixels : int, optional
        If this parameter is passed, the algoritm will stop searching
        for candidate pixels after reaching a count equal to this value,
        by default None.
    """

    def __init__(self, lower_bound=(0, 0, 0), upper_bound=(255, 255, 255),
                 color_space='BGR', max_pixels=None):
        super().__init__()
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.color_space = color_space
        self.max_pixels = max_pixels

    def detect(self, frame, roi_bound=None, preprocessing=None):
        """
        Identifies the tracked object in the image ``frame``
        by thresholding it using the bound parameters defined when
        the object was constructed.

        Parameters
        ----------
        frame : np.ndarray
            Image containing the object to be tracked
        roi_bound : tuple, optional
            Coordinates of the region of interest of the frame. The
            expected format if a tuple with the form (xmin, xmax, ymin,
            ymax). If passed the algorithm will crop this region of the
            frame and will proceed only in this region, providing the
            estimations refered to this region instead of the whole
            image, by default None.
        preprocessing : func
            A function to be applied to the frame (Or cropped version
            of it if roi_bound is passed) before detecting the object
            on it, by default None.

        Returns
        -------
        np.ndarray
            A binary version of ``frame`` where elements with value
            ``0`` indicate the absence of object and ``1`` the precense
            of the object.
        tuple
            x, y coordinates of the centroid of the object in the image.
        """

        # Make a preprocessed (and copied) version of the frame
        frame = self.preprocess(frame, roi_bound, preprocessing)

        # Convert image to desired colorspace
        copied_image = _change_colorspace(frame, self.color_space)

        mask = cv2.inRange(copied_image, self.lower_bound, self.upper_bound)
        centroid = self.get_centroid(mask)
        # Convert the grayscale image to binary image
        return mask, centroid


class FrameDifferencing(TrackingAlgorithm):
    """
    Identifies the position of an object by comparison between
    consecutive frames

    Parameters
    ----------
        Minimum difference (in terms of pixel intensity) among current
        and previous image to consider that pixel as part of a moving
        object, by default 1.
    """

    def __init__(self, frame_diff_threshold=1):
        super().__init__()
        self.frame_diff_threshold = frame_diff_threshold
        self.prev_frame = None

    def detect(self, frame, roi_bound=None, preprocessing=None):
        """
        Identifies the tracked object in the image ``frame``
        by comparing the difference with the previous frames. All the
        pixels differing by more than frame_diff_threshold will be
        considered part of the moving object.

        Parameters
        ----------
        frame : np.ndarray
            Image containing the object to be tracked
        roi_bound : tuple, optional
            Coordinates of the region of interest of the frame. The
            expected format if a tuple with the form (xmin, xmax, ymin,
            ymax). If passed the algorithm will crop this region of the
            frame and will proceed only in this region, providing the
            estimations refered to this region instead of the whole
            image, by default None.
        preprocessing : func
            A function to be applied to the frame (Or cropped version
            of it if roi_bound is passed) before detecting the object
            on it, by default None.

        Returns
        -------
        np.ndarray
            A binary version of ``frame`` where elements with value
            ``0`` indicate the absence of object and ``1`` the precense
            of the object.
        tuple
            x, y coordinates of the centroid of the object in the image.
        """

        if self.prev_frame is None:
            self.prev_frame = frame.copy()

        # Make a preprocessed (and copied) version of the frame
        cframe = self.preprocess(frame, roi_bound, preprocessing)
        prev_frame = self.preprocess(self.prev_frame, roi_bound, preprocessing)

        diff = cv2.absdiff(cframe, prev_frame)

        # Convert image to grayscale image
        gray_image = cv2.cvtColor(diff, cv2.COLOR_BGR2GRAY)

        # Convert the grayscale image to binary image
        mask = cv2.inRange(gray_image, self.frame_diff_threshold, 255)

        # Compute the centroid of the pixels over threshold
        centroid = self.get_centroid(mask)

        # Update previous image cache
        self.prev_frame = frame.copy()

        # Convert the grayscale image to binary image
        return mask, centroid


class BackgroundSubtraction(TrackingAlgorithm):
    """
    Identifies the position of an object by subtracting a known
    background.

    Parameters
    ----------
    background : np.ndarray
        Image containing the actual background of the scene where the
        images were taken. This algorithm will detect as an object of
        interest everything that differs from the background.
    background_threshold : int, optional
        Minimum difference (in terms of pixel intensity) among current
        image and background to consider that pixel as part of a moving
        object, by default 1.
    """

    def __init__(self, background, background_threshold):
        super().__init__()
        self.background_threshold = background_threshold
        self.background = background

    def detect(self, frame, roi_bound=None, preprocessing=None):
        """
        Identifies the tracked object in the image ``frame``
        by comparing the difference with the background. All the pixels
        differing by more than background_threshold will be considered
        part of the moving object.

        Parameters
        ----------
        frame : np.ndarray
            Image containing the object to be tracked
        roi_bound : tuple, optional
            Coordinates of the region of interest of the frame. The
            expected format if a tuple with the form (xmin, xmax, ymin,
            ymax). If passed the algorithm will crop this region of the
            frame and will proceed only in this region, providing the
            estimations refered to this region instead of the whole
            image, by default None.
        preprocessing : func
            A function to be applied to the frame (Or cropped version
            of it if roi_bound is passed) before detecting the object
            on it, by default None.

        Returns
        -------
        np.ndarray
            A binary version of ``frame`` where elements with value
            ``0`` indicate the absence of object and ``1`` the precense
            of the object.
        tuple
            x, y coordinates of the centroid of the object in the image.
        """

        # Make a preprocessed (and copied) version of the frame
        cframe = self.preprocess(frame, roi_bound, preprocessing)
        backgrn_roi = self.preprocess(self.background, roi_bound, preprocessing)

        diff = cv2.absdiff(cframe, backgrn_roi)

        # Convert image to grayscale image
        gray_image = cv2.cvtColor(diff, cv2.COLOR_BGR2GRAY)

        # Convert the grayscale image to binary image
        mask = cv2.inRange(gray_image, self.background_threshold, 255)

        # Compute the centroid of the pixels over threshold
        centroid = self.get_centroid(mask)

        # Convert the grayscale image to binary image
        return mask, centroid


class TemplateMatching(TrackingAlgorithm):
    """
    Identifies the position of an object by correlating with a template.

    Parameters
    ----------
    template : np.ndarray
        Image containing a template of a tipical image of the object
        being tracked. This algorithm will detect as an object of
        interest the point with higher correlation between the template
        and the image.
    threshold : float, optional
        Minimum value of correlation to be considered as a match, by
        default 0.8.
    """

    def __init__(self, template, threshold):
        super().__init__()
        self.template = cv2.cvtColor(template, cv2.COLOR_BGR2GRAY)
        self.threshold = threshold

        self.w, self.h = self.template.shape[::-1]

    def detect(self, frame, roi_bound=None, preprocessing=None):
        """
        Identifies the tracked object in the image ``frame``
        by comparing each region with a template. The region with higher
        correlation will be selected as the current position of the
        object.

        Parameters
        ----------
        frame : np.ndarray
            Image containing the object to be tracked
        roi_bound : tuple, optional
            Coordinates of the region of interest of the frame. The
            expected format if a tuple with the form (xmin, xmax, ymin,
            ymax). If passed the algorithm will crop this region of the
            frame and will proceed only in this region, providing the
            estimations refered to this region instead of the whole
            image, by default None.
        preprocessing : func
            A function to be applied to the frame (Or cropped version
            of it if roi_bound is passed) before detecting the object
            on it, by default None.

        Returns
        -------
        np.ndarray
            A binary version of ``frame`` where elements with value
            ``0`` indicate the absence of object and ``1`` the precense
            of the object.
        tuple
            x, y coordinates of the centroid of the object in the image.
        """

        # Make a preprocessed (and copied) version of the frame
        cframe = self.preprocess(frame, roi_bound, preprocessing)

        # Convert image to grayscale image
        gray_img = cv2.cvtColor(cframe, cv2.COLOR_BGR2GRAY)

        # Compute the correlation
        res = cv2.matchTemplate(gray_img, self.template, cv2.TM_CCOEFF_NORMED)

        # Get the point of max correlation
        pt = np.unravel_index(res.argmax(), res.shape)

        # Compute the centroid of the region with max correlation
        centroid = None
        if res[pt] > self.threshold:
            centroid = (int(pt[1] + self.w/2), int(pt[0] + self.h/2))


        # Convert the grayscale image to binary image
        mask = None

        # Convert the grayscale image to binary image
        return mask, centroid


class OpticalFlow(TrackingAlgorithm):
    """
    This class implements optical flow based on
    Gunner Farneback's algorithm. A section of the
    frame is selected and tracked using dense optical flow.

    Parameters
    ----------
    threshold : float
        Minimum value for the magnitude of optical flow to be considered
        part of the motion.
    buffer_size : int, optional
        Indicates how many frames in the past the algorithm is going to
        look before computing the optical flow, by default 1.
    """

    def __init__(self, threshold, buffer_size=1):
        super().__init__()
        self.threshold = threshold
        self.previous_frames = []

        assert buffer_size > 0

        self.buffer_size = buffer_size


    def detect(self, frame, roi_bound=None, preprocessing=None):
        """
        Identifies the tracked object in the image ``frame``
        by tracking the motion of a region using optical flow.

        Parameters
        ----------
        frame : np.ndarray
            Image containing the object to be tracked
        roi_bound : tuple, optional
            Coordinates of the region of interest of the frame. The
            expected format if a tuple with the form (xmin, xmax, ymin,
            ymax). If passed the algorithm will crop this region of the
            frame and will proceed only in this region, providing the
            estimations refered to this region instead of the whole
            image, by default None.
        preprocessing : func
            A function to be applied to the frame (Or cropped version
            of it if roi_bound is passed) before detecting the object
            on it, by default None.

        Returns
        -------
        np.ndarray
            A binary version of ``frame`` where elements with value
            ``0`` indicate the absence of object and ``1`` the precense
            of the object.
        tuple
            x, y coordinates of the centroid of the object in the image.
        """

        if len(self.previous_frames) == self.buffer_size:

            cframe = self.preprocess(
                frame=frame,
                roi_bound=roi_bound,
                preprocessing=preprocessing
            )
            pframe = self.preprocess(
                frame=self.previous_frames[-1],
                roi_bound=roi_bound,
                preprocessing=preprocessing
            )

            cframe = cv2.cvtColor(cframe, cv2.COLOR_BGR2GRAY)
            pframe = cv2.cvtColor(pframe, cv2.COLOR_BGR2GRAY)

            diff = cv2.calcOpticalFlowFarneback(
                prev=pframe,
                next=cframe,
                flow=None,
                pyr_scale=0.5,
                levels=3,
                winsize=15,
                iterations=3,
                poly_n=5,
                poly_sigma=1.2,
                flags=0
            )
            mag, _ = cv2.cartToPolar(diff[..., 0], diff[..., 1])

            # Convert the grayscale image to binary image
            mask = cv2.inRange(mag, self.threshold, 255)

            # Compute the centroid of the pixels over threshold
            centroid = self.get_centroid(mask)

            self.previous_frames.pop(0)
        else:
            mask, centroid = None, None

        # Store the current frame
        self.previous_frames.append(frame.copy())

        return mask, centroid



import cv2
import nudged
import numpy as np

# ShiTomasi corner detection
feature_params = dict(
    maxCorners=30,
    qualityLevel=0.6,
    minDistance=30,
    blockSize=100
)

# Lucas Kanade optical flow
lk_params = dict(
    winSize=(20, 20),
    maxLevel=15,
    criteria=(cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT, 200, 0.05)
)


def _rot_matrix(theta, inv=False):
    R = np.array([[np.cos(theta), -np.sin(theta)],
                  [np.sin(theta), np.cos(theta)]])
    if inv:
        R = R.T
    return R


def _affine_matrix(theta, tx, ty, scale=1, R_inv=False):
    R = _rot_matrix(theta, R_inv)
    R_s = (scale * np.identity(2)) @ R
    T = np.array([tx, ty])
    A = np.hstack([R_s, T[None, :].T])
    return A


def _estimate_params(p1, p2):
    trans = nudged.estimate(p1, p2)
    tx, ty = trans.get_translation()
    theta, scale = trans.get_rotation(), trans.get_scale()
    return theta, tx, ty, scale


def _get_p3(p1, theta, tx, ty, scale=1):
    A = _affine_matrix(theta, tx, ty, scale)
    p3 = np.array([A @ [x1, y1, 1] for (x1, y1) in p1])
    return p3


def _get_r(p1, p2):
    x, y = (p1 - p2).T
    r = np.sqrt(x**2 + y**2)
    return r


def _get_mask_r(r, quantile=1.5):
    # Median and standard deviation of distance population
    r_median, r_std = np.median(r), np.std(r)
    # Cutoff used to filter far features points
    bound = max(1, quantile * r_std)
    # Get mask for nearby points
    mask = r < r_median + bound
    return mask


def _delete_far_points(p1, p2, p3=None, quantile=1.5):
    # Distance amoung p1 and p2, or p2 and p3
    r = _get_r(p1, p2) if p3 is None else _get_r(p2, p3)
    # Mask that filters outliers for a given quantile value
    mask = _get_mask_r(r, quantile)
    # Delete far points
    p1, p2 = p1[mask], p2[mask]
    return p1, p2


def _estimate_matrix(p1, p2, quantile1=None, quantile2=None):
    # Validate tracked features deletting outliers
    if quantile1 is not None:
        p1, p2 = _delete_far_points(p1, p2, quantile=quantile1)

    # Estimate matrix parameters and transformed features
    affine_params = _estimate_params(p1, p2)
    p3 = _get_p3(p1, *affine_params)

    # Delete outliers considering transformed features
    if quantile2 is not None:
        p1, p2 = _delete_far_points(p1, p2, p3, quantile=quantile2)
        affine_params = _estimate_params(p1, p2)
        p3 = _get_p3(p1, *affine_params)

    p = p1, p2, p3
    return p, affine_params


def _get_mse(p2, p3):
    x, y = (p2 - p3).T
    r_2 = np.mean(x**2 + y**2)
    mse_r = np.sqrt(r_2)

    x_2, y_2 = np.mean(x**2), np.mean(y**2)
    mse_x, mse_y = np.sqrt(x_2), np.sqrt(y_2)

    return mse_r, mse_x, mse_y


def _get_affine(img1, img2, region, mask=None):
    x0, xf, y0, yf = region
    err = None

    # Get main regions
    img1_ = img1[y0:yf, x0:xf, :]
    img2_ = img2[y0:yf, x0:xf, :]

    if mask is not None:
        mask = mask[y0:yf, x0:xf]

    # Convert to grayscale
    img1_gray = cv2.cvtColor(img1_, cv2.COLOR_BGR2GRAY)
    img2_gray = cv2.cvtColor(img2_, cv2.COLOR_BGR2GRAY)

    # Equalize histograms to improve contrast
    img1_gray = cv2.equalizeHist(img1_gray)
    img2_gray = cv2.equalizeHist(img2_gray)

    # Track good features
    p1_ = cv2.goodFeaturesToTrack(img1_gray, mask=mask, **feature_params)
    p2_, st, _ = cv2.calcOpticalFlowPyrLK(img1_gray, img2_gray, p1_,
                                          None, **lk_params)

    # Change origin and select tracked points
    p1, p2 = p1_ + [x0, y0], p2_ + [x0, y0]
    p1_good, p2_good = p1[st == 1], p2[st == 1]

    # Cancel estimation if no good points were found or tracked
    if p1_good.size == 0:
        print('[ERROR] No good points were found or sucessfully tracked.')
        return 3*(0,), 3*(0,), err

    # Estimate points and matrix
    p_good, affine_params = _estimate_matrix(p1_good, p2_good,
                                            quantile1=1.5, quantile2=1.5)
    p1_good, p2_good, p3_good = p_good

    # Mean square error
    err, *_ = _get_mse(p2_good, p3_good)

    return p_good, affine_params, err



"""
This module is designed to capture Trajectory objects from
a sequence of images. It contains different tools to extract
the trajectories based on the color or the motion of the object
being tracked.

All the resources of this module should be imported directly
from ``yupi.tracking``.
"""

from yupi.tracking.trackers import (
    ROI,
    ObjectTracker,
    CameraTracker,
    TrackingScenario
)
from yupi.tracking.algorithms import (
    TrackingAlgorithm,
    ColorMatching,
    FrameDifferencing,
    BackgroundSubtraction,
    TemplateMatching,
    OpticalFlow,
    BackgroundEstimator
)
from yupi.tracking.undistorters import (
    Undistorter,
    ClassicUndistorter,
    RemapUndistorter
)

__all__ = [
    'ROI',
    'ObjectTracker',
    'CameraTracker',
    'TrackingScenario',
    'TrackingAlgorithm',
    'FrameDifferencing',
    'ColorMatching',
    'BackgroundSubtraction',
    'TemplateMatching',
    'OpticalFlow',
    'BackgroundEstimator',
    'Undistorter',
    'ClassicUndistorter',
    'RemapUndistorter',
]



"""
This module contains different statistical models to
generate trajectories given certain statistical constrains.

All the resources of this module should be imported directly
from ``yupi.generators``.
"""

from yupi.generators._generators import (
    Generator,
    RandomWalkGenerator,
    LangevinGenerator,
    DiffDiffGenerator
)

__all__ = [
    'Generator',
    'RandomWalkGenerator',
    'LangevinGenerator',
    'DiffDiffGenerator'
]



import abc
import numpy as np
import cv2


class Undistorter(metaclass=abc.ABCMeta):
    """
    Abstract class to model an undistortion method to be aplied on
    images in order to correct the spherical distortion caused by
    the camera lens. Classes inheriting from this class should
    implement ``undistort`` method.

    To use an undistorion method you will need to obtain the calibration
    matrix of your camera. You can follow the guide in opencv docs
    until the end of the section ``Undistortion`` to compute the matrix
    of your camera:

    https://docs.opencv.org/master/dc/dbb/tutorial_py_calibration.html

    Then you can save the matrix and other parameters in an npz file
    using numpy:

    >>> np.savez("camera_file.npz", h=h, w=w, mtx=mtx, dist=dist, newcameramtx=newcameramtx)

    Parameters
    ----------
    camera_file : str
        Path to the camera calibration file ("camera_file.npz" in the
        above example).
    turn : bool
        This parameter is used to rotate 90 degrees the frame, before
        undistorting it. It is useful when the input video is rotated
        respect the orginal orientation used when the camera was
        calibrated (Not a very frequent use case). The undistorted
        result will be rotated -90 degrees before returning. By default
        is False.
    """

    def __init__(self, camera_file, turn=False):
        # Read camera undistort matrix
        npzfile = np.load(camera_file)

        # Initialize camera parameters
        self.c_h = npzfile['h']
        self.c_w = npzfile['w']
        self.c_mtx = npzfile['mtx']
        self.c_dist = npzfile['dist']
        self.c_newcameramtx = npzfile['newcameramtx']
        self.turn = turn
        c_map = cv2.initUndistortRectifyMap(
            cameraMatrix=self.c_mtx,
            distCoeffs=self.c_dist,
            R=None,
            newCameraMatrix=self.c_newcameramtx,
            size=(self.c_w, self.c_h),
            m1type=5
        )
        self.c_mapx, self.c_mapy = c_map
        self.mask = None
        self.background = None

    @abc.abstractmethod
    def undistort(self, frame):
        """
        Abstract method that is implemented on inheriting classes. It
        should compute an undistorted version of frame using the given
        camera calibration matrix and a method specific to the
        inheriting class.
        """

    # Turn the image if required
    def _rotate(self, frame, _input=True):
        if self.turn:
            direction = cv2.ROTATE_90_COUNTERCLOCKWISE
            if _input:
                direction = cv2.ROTATE_90_CLOCKWISE
            frame = cv2.rotate(frame, direction)
        return frame

    # Method to be called to fix the distortion
    def fix(self, frame):
        frame = self._rotate(frame, _input=True)
        if self.mask is None:
            self._create_mask(frame)
        corrected = self.undistort(frame)
        return self.masked(corrected)

    # Create a mask with the distortion pattern
    def _create_mask(self, frame):
        empty_frame = 255 * np.ones(frame.shape, dtype=np.uint8)
        corrected = self.undistort(empty_frame)
        kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (5, 5))
        self.mask = cv2.erode(corrected, kernel, iterations=1)
        self.mask = cv2.bitwise_not(self.mask)
        self.background = np.full(frame.shape, 130, dtype=np.uint8)

    # Apply the mask to a frame to adjust border colors
    def masked(self, frame):
        frame = cv2.bitwise_or(frame, self.mask)
        return  self._rotate(frame, _input=False)


class ClassicUndistorter(Undistorter):
    """
    Undistorter that performs undistortion using ``cv2.undistort``.

    Parameters
    ----------
    camera_file : str
        Path to the camera calibration file ("camera_file.npz" in the
        above example).
    turn : bool
        This parameter is used to rotate 90 degrees the frame, before
        undistorting it. It is useful when the input video is rotated
        respect the orginal orientation used when the camera was
        calibrated (Not a very frequent use case). The undistorted
        result will be rotated -90 degrees before returning. By default
        is False.
    """

    def undistort(self, frame):
        """
        Computes the undistorted version of ``frame`` using
        ``cv2.undistort``.

        Returns
        ----------
        np.ndarray
            Undistorted version of frame.
        """

        return cv2.undistort(
            src=frame,
            cameraMatrix=self.c_mtx,
            distCoeffs=self.c_dist,
            dst=None,
            newCameraMatrix=self.c_newcameramtx
        )


class RemapUndistorter(Undistorter):
    """
    Undistorter that performs undistortion using ``cv2.remap``.

    Parameters
    ----------
    camera_file : str
        Path to the camera calibration file ("camera_file.npz" in the
        above example).
    turn : bool
        This parameter is used to rotate 90 degrees the frame, before
        undistorting it. It is useful when the input video is rotated
        respect the orginal orientation used when the camera was
        calibrated (Not a very frequent use case). The undistorted
        result will be rotated -90 degrees before returning. By default
        is False.
    """

    def undistort(self, frame):
        """
        Computes the undistorted version of ``frame`` using
        ``cv2.remap``.

        Returns
        ----------
        np.ndarray
            Undistorted version of frame.
        """

        return cv2.remap(frame, self.c_mapx, self.c_mapy, cv2.INTER_LINEAR)



"""
All the resources of the library should be imported directly
from one of the aforementioned modules.
"""

import logging
from yupi.features import Features
from yupi.trajectory import Trajectory, TrajectoryPoint
from yupi.vector import Vector

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s [%(levelname)s] %(message)s',
    datefmt='%H:%M:%S'
)

__all__ = [
    'Trajectory',
    'TrajectoryPoint',
    'Features',
    'Vector'
]

__version__ = '0.8.0'



from typing import List
import numpy as np


class Features():
    """
    Extracts useful information from a trajectory.

    Parameters
    ----------
    traj : Trajectory
        Input trajectory
    """

    def __init__(self, traj):
        self.traj = traj

    @property
    def mean_vel(self) -> float:
        """float : Calculates de mean velocity"""
        return float(np.average(self.traj.v))

    @property
    def displacement(self) -> float:
        """float : Calculates de displacement"""
        traj = self.traj
        return (traj.r[-1] - traj.r[0]).norm

    @property
    def length(self) -> float:
        """float : Calculates de length"""
        traj = self.traj
        return sum(traj.delta_r.norm)

    def as_dict(self, only: List[str] = None,
                remove: List[str] = None) -> dict:
        """
        Get all the features as dictionary.

        Parameters
        ----------
        only : List[str], optional
            Filters the features by giving the ones named in the list,
            by default None.
        remove : List[str], optional
            Filters the features by removing the ones named in the list,
            by default None.

        Returns
        -------
        dict
            Features dictionary.
        """

        _dict = dict(self.__class__.__dict__)
        _dict = {k:v for k,v in _dict.items() if isinstance(v, property)}
        _keys = _dict.keys()
        if remove is not None:
            _keys = [k for k in _keys if k not in remove]
        if only is not None:
            _keys = [k for k in _keys if k in only]
        _dict = {k:v.__get__(self) for k,v in _dict.items() if k in _keys}
        return _dict





class YupiException(Exception):
    """Generic exception for yupi package"""

class TrajectoryError(YupiException):
    """Generic exception for handling trajectory errors"""

class LoadTrajectoryError(TrajectoryError):
    """Error while loading a trajectory"""

    def __init__(self, path):
        self.path = path
        self.message = f"File: '{self.path}' is not a valid trajectory"
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message



import numpy as np
from yupi.trajectory import Trajectory


def exp_convolutional_filter(traj: Trajectory, gamma: float,
                             new_traj_id: str = None):
    """
    Returns a smoothed version of the trajectory `traj`
    by taking a weighted average over past values.

    Parameters
    ----------
    traj : Trajectory
        Input trajectory.
    gamma : float
        Inverse of the characteristic time window of
        the average.

    Returns
    -------
    Trajectory
        Smoothed trajectory.
    """

    r = traj.r
    dt = np.ediff1d(traj.t)
    R = np.zeros_like(r)
    for i in range(len(traj) - 1):
        R[i + 1] = R[i] - gamma * (R[i] - r[i]) * dt[i]

    smooth_traj = Trajectory(points=R, t=traj.t, traj_id=new_traj_id)
    return smooth_traj



import numpy as np
from numpy.linalg.linalg import norm as nrm


class Vector(np.ndarray):
    """Represents a vector"""

    @property
    def norm(self):
        """Vector : Calculates the norm of the vector. If the vector
        is alist of vectors then the norm of each item is calculated"""
        if len(self.shape) < 2:
            return nrm(self)
        return Vector.create([nrm(p) for p in self])

    @property
    def delta(self):
        """Vector : Calculates the differnece between each item"""
        if len(self.shape) > 1:
            new_vec = []
            for i in range(self.shape[1]):
                new_vec.append(np.ediff1d(self[:, i]))
            return Vector.create(new_vec).T
        else:
            return Vector.create(np.ediff1d(self))

    @property
    def x(self):
        """Vector : X component of all vector items"""
        return self.component(0)

    @property
    def y(self):
        """Vector : Y component of all vector items"""
        return self.component(1)

    @property
    def z(self):
        """Vector : Z component of all vector items"""
        return self.component(2)

    def component(self, dim):
        """
        Extract a given component from all vector items.

        Parameters
        ----------
        dim : int
            Component index.

        Returns
        -------
        Vector
            Component extracted.

        Raises
        ------
        TypeError
            If the vector has no axis 1.
        TypeError
            If `dim` is not an integer.
        ValueError
            If the shape of axis 1 of the vector is lower than dim.

        Examples
        --------
        >>> v = Vector.create([[1,2],[0,2],[3,0]])
        >>> v.component(0)
        Vector([1, 0, 3])
        >>> v.component(1)
        Vector([2, 2, 0])
        """

        if len(self.shape) < 2:
            raise TypeError('Operation not supperted on simple vectors')
        if not isinstance(dim, int):
            raise TypeError("Parameter 'dim' must be an integer")
        if self.shape[1] < dim + 1:
            raise ValueError(f'Vector has not component {dim}')
        return self[:, dim]

    @staticmethod
    def create(*args, **kwargs):
        """
        Creates a new vector.

        Returns
        -------
        Vector
            Vector created
        """

        arr = np.array(*args, **kwargs)
        return arr.view(Vector)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['yupi',
 'yupi.generators',
 'yupi.graphics',
 'yupi.stats',
 'yupi.tracking',
 'yupi.transformations']

package_data = \
{'': ['*']}

install_requires = \
['PyQt5>=5.15.3,<6.0.0',
 'matplotlib>=3.2.0',
 'nudged>=0.3.1',
 'numpy>=1.16.5',
 'opencv-python>=4.4.0']

setup_kwargs = {
    'name': 'yupi',
    'version': '0.8.0',
    'description': 'A package for tracking and analysing objects trajectories',
    'long_description': '<p style="text-align:center;"><img src="logo.png" alt="Logo"></p>\n\nStanding for *Yet Underused Path Instruments*, **yupi** is a set of tools designed\nfor collecting, generating and processing trajectory data.\n\n## Installation\n\nCurrent recommended installation method is via the pypi package:\n\n```cmd\npip install yupi\n```\n\n## Getting Started\n\nIn the [official documentation](https://yupi.readthedocs.io/en/latest/) there\nare some resources to start using the library: Tutorials, Examples  and a\ndetailed description of the API.\n\n## Examples\n\nCode examples (with additional multimedia resources) can be found in\n[this repository](https://github.com/yupidevs/yupi_examples). Additionally, in\nthe [Examples section](https://yupi.readthedocs.io/en/latest/examples/examples.html)\nof the documentation, you can find the same examples with additional comments\nand expected execution results in order to inspect the examples without actually\nexecuting them.\n',
    'author': 'Gustavo Viera-López',
    'author_email': 'gvieralopez@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
This module contains a set of functions for statistical information
extraction from a trajectory collection.

All the resources of this module should be imported directly
from ``yupi.stats``.
"""

from yupi.stats._stats import (
    collect_at,
    collect,
    turning_angles_ensemble,
    speed_ensemble,
    msd_ensemble,
    msd_time,
    vacf_ensemble,
    msd,
    vacf_time,
    vacf,
    kurtosis,
    kurtosis_reference,
    psd
)

__all__ = [
    'turning_angles_ensemble',
    'collect_at',
    'collect',
    'speed_ensemble',
    'msd_ensemble',
    'msd_time',
    'vacf_ensemble',
    'msd',
    'vacf_time',
    'vacf',
    'kurtosis',
    'kurtosis_reference',
    'psd'
]



"""
This module constains a set of functions dadicated to plot
trajectories and extracted statistical data from ``yupi.stats``

All the resources of this module should be imported directly
from ``yupi.graphics``.
"""

from yupi.graphics._stats_plots import (
    plot_angles_hist,
    plot_kurtosis,
    plot_msd,
    plot_vacf,
    plot_velocity_hist,
    plot_psd,
    plot_hist,
    plot_hists
)

from yupi.graphics._trajs_plots import plot_2D, plot_3D

from yupi.graphics._style import (
    LINE,
    LINE_DOTTED,
    DOTTED
)

__all__ = [
    'plot_angles_hist',
    'plot_kurtosis',
    'plot_msd',
    'plot_vacf',
    'plot_velocity_hist',
    'plot_2D',
    'plot_3D',
    'plot_psd',
    'plot_hist',
    'plot_hists',
    'LINE',
    'LINE_DOTTED',
    'DOTTED'
]
